# -*- coding: utf-8 -*-


import types

from flask import Flask, render_template
from flask_cors import CORS
from pathlib import Path

from rekcurd_dashboard.utils import RekcurdDashboardConfig
from rekcurd_dashboard.models import db
from rekcurd_dashboard.apis import api
from rekcurd_dashboard.auth import auth


def create_app(config_file: str = None, logger_file: str = None, **options) -> (Flask, RekcurdDashboardConfig):
    """create_app."""

    app = Flask(__name__, static_folder='static', template_folder='static/dist')
    app.logger.disabled = True
    for h in app.logger.handlers[:]:
        app.logger.removeHandler(h)

    @app.route('/ui/')
    @app.route('/ui/<path:path>')
    def root_url(**kwargs):
        return render_template('index.html')

    # load configurations
    config = RekcurdDashboardConfig(config_file)
    config.set_configurations(**options)
    kubedir = Path(config.DIR_KUBE_CONFIG)
    if not kubedir.exists():
        kubedir.mkdir(parents=True, exist_ok=True)

    # set configurations
    app.config['SQLALCHEMY_DATABASE_URI'] = config.SQLALCHEMY_DATABASE_URI
    app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = True
    app.config['DEBUG'] = config.DEBUG_MODE
    app.config['SWAGGER_UI_DOC_EXPANSION'] = 'list'
    app.config['RESTPLUS_VALIDATE'] = True
    app.config['MAX_CONTENT_LENGTH'] = 1073741824

    # logger
    try:
        l = types.ModuleType('logger')
        with open(logger_file, mode='rb') as f:
            exec(compile(f.read(), logger_file, 'exec'), l.__dict__)
        logger = l.__dict__.get("logger")
        logger.info("Use custom logger.")
    except Exception as e:
        from rekcurd_dashboard.logger import logger
        logger.info(str(e))
        logger.info("Invalid logger.")
        logger.info("Use default logger.")

    # initialize applications
    api.init_app(app, dashboard_config=config, logger=logger)
    auth.init_app(app, api, config.AUTH_CONFIG, logger=logger)
    CORS(app)
    db.init_app(app)
    db.create_all(app=app)

    return app, config


def main(args) -> None:
    app, config = create_app(*args[1:])
    app.run(host='0.0.0.0', port=config.SERVICE_PORT, threaded=True)


if __name__ == '__main__':
    import sys
    main(sys.argv)



from flask_restplus import Namespace, fields, Resource, reqparse

from . import status_model, load_istio_routing, apply_new_route_weight
from rekcurd_dashboard.models import db, KubernetesModel, ApplicationModel, ServiceModel
from rekcurd_dashboard.utils import RekcurdDashboardException


service_routing_api_namespace = Namespace('service_routings', description='Service Routing API Endpoint.')
success_or_not = service_routing_api_namespace.model('Success', status_model)
service_weight_params = service_routing_api_namespace.model('Weight', {
    'display_name': fields.String(
        readOnly=True,
        description='Service name.'
    ),
    'service_id': fields.String(
        readOnly=True,
        description='Service ID.'
    ),
    'service_weight': fields.Integer(
        readOnly=True,
        description='Service weight.'
    )
})
service_routing_params = service_routing_api_namespace.model('Routing', {
    'application_name': fields.String(
        readOnly=True,
        description='Application name.'
    ),
    'service_level': fields.String(
        readOnly=True,
        description='Service level. [development/beta/staging/sandbox/production]',
        example='development'
    ),
    'service_weights': fields.List(
        fields.Nested(service_weight_params),
        readOnly=True,
        description='Service weights.'
    )
})
service_weight_expect = service_routing_api_namespace.model('Weight', {
    'service_id': fields.String(
        required=True,
        description='Service ID.'
    ),
    'service_weight': fields.Integer(
        required=True,
        description='Service weight.'
    )
})
service_routing_expect = service_routing_api_namespace.model('Routing', {
    'service_level': fields.String(
        required=True,
        description='Service level. [development/beta/staging/sandbox/production]',
        example='development'
    ),
    'service_weights': fields.List(
        fields.Nested(service_weight_expect),
        required=True,
        description='Service weights.'
    )
})


@service_routing_api_namespace.route('/projects/<int:project_id>/applications/<application_id>/service_routing')
class ApiServiceRouting(Resource):
    get_parser = reqparse.RequestParser()
    get_parser.add_argument(
        'service_level', location='args', type=str, required=True,
        choices=('development', 'beta', 'staging', 'sandbox', 'production'),
        help='Service level. [development/beta/staging/sandbox/production].')

    service_routing_parser = reqparse.RequestParser()
    service_routing_parser.add_argument(
        'service_level', type=str, required=True,
        choices=('development','beta','staging','sandbox','production'),
        help='Service level. [development/beta/staging/sandbox/production].')
    service_routing_parser.add_argument(
        'service_weights', type=dict, action='append', required=True,
        help='Service weights.')

    @service_routing_api_namespace.marshal_with(service_routing_params)
    @service_routing_api_namespace.expect(get_parser)
    def get(self, project_id: int, application_id: str):
        """Get routing info."""
        args = self.get_parser.parse_args()
        service_level = args["service_level"]
        kubernetes_model: KubernetesModel = db.session.query(KubernetesModel).filter(
            KubernetesModel.project_id == project_id).first()
        application_model: ApplicationModel = db.session.query(ApplicationModel).filter(
            ApplicationModel.application_id == application_id).first_or_404()
        service_models = db.session.query(ServiceModel).filter(
            ServiceModel.application_id == application_id,
            ServiceModel.service_level == service_level).all()
        if not service_models:
            raise RekcurdDashboardException("No services available.")

        service_id_name = dict()
        for service_model in service_models:
            service_id_name[service_model.service_id] = service_model.display_name
        routes = load_istio_routing(kubernetes_model, application_model, service_level)
        response_body = dict()
        response_body["application_name"] = application_model.application_name
        response_body["service_level"] = service_level
        service_weights = list()
        response_body["service_weights"] = service_weights
        for route in routes:
            service_id = route["destination"]["host"][4:]
            display_name = service_id_name.pop(service_id)
            weight = route["weight"]
            service_weights.append({
                "display_name": display_name,
                "service_id": service_id,
                "service_weight": weight})
        for service_id, display_name in service_id_name.items():
            service_weights.append({
                "display_name": display_name,
                "service_id": service_id,
                "service_weight": 0})
        return response_body

    @service_routing_api_namespace.marshal_with(success_or_not)
    @service_routing_api_namespace.expect(service_routing_expect)
    def patch(self, project_id: int, application_id: str):
        """Update routing weights."""
        service_routing_args = self.service_routing_parser.parse_args()
        service_level: str = service_routing_args["service_level"]
        service_ids = []
        service_weights = []
        for service_weight in service_routing_args["service_weights"]:
            service_ids.append(service_weight["service_id"])
            service_weights.append(service_weight["service_weight"])
        apply_new_route_weight(project_id, application_id, service_level, service_ids, service_weights)
        return {"status": True, "message": "Success."}



import json
import uuid

from datetime import datetime
from pathlib import Path

from . import api, RekcurdDashboardException
from .common import kubernetes_cpu_to_float
from rekcurd_dashboard.models import db, DataServerModel, KubernetesModel, ApplicationModel, ServiceModel, ModelModel


def get_full_config_path(filename: str):
    """
    Get full path of Kubernetes config.
    :param filename:
    :return:
    """
    return f'{api.dashboard_config.DIR_KUBE_CONFIG}/{filename}'


def save_kubernetes_access_file(file, config_path):
    """
    Save Kubernetes config.
    :param file:
    :param config_path:
    :return:
    """
    full_config_path = get_full_config_path(config_path)
    file.save(full_config_path)
    return


def remove_kubernetes_access_file(config_path):
    """
    Remove Kubernetes config.
    :param config_path:
    :return:
    """
    full_config_path = get_full_config_path(config_path)
    Path(full_config_path).unlink()
    return


def update_kubernetes_deployment_info(kubernetes_model: KubernetesModel):
    """
    Update Kubernetes deployment info.
    :param kubernetes_model:
    :return:
    """
    full_config_path = get_full_config_path(kubernetes_model.config_path)
    from kubernetes import client, config
    config.load_kube_config(full_config_path)
    v1_api = client.AppsV1Api()
    list_deployment_for_all_namespaces = v1_api.list_deployment_for_all_namespaces(watch=False)

    """Application registration."""
    for i in list_deployment_for_all_namespaces.items:
        labels = i.metadata.labels
        if labels is None or labels.get("rekcurd-worker", "False") == "False":
            continue

        application_id = labels["id"]
        application_name = labels["name"]
        application_model = db.session.query(ApplicationModel).filter(
            ApplicationModel.application_id == application_id).one_or_none()
        if application_model is None:
            application_model = ApplicationModel(
                project_id=kubernetes_model.project_id,
                application_id=application_id,
                application_name=application_name)
            db.session.add(application_model)
            db.session.flush()

    """Service registration."""
    for i in list_deployment_for_all_namespaces.items:
        labels = i.metadata.labels
        if labels is None or labels.get("rekcurd-worker", "False") == "False":
            continue

        application_id = labels["id"]
        service_id = labels["sel"]
        service_model: ServiceModel = db.session.query(ServiceModel).filter(
            ServiceModel.service_id == service_id).one_or_none()
        if service_model is None:
            service_level = i.metadata.namespace
            version = None
            filepath = None
            insecure_host = None
            insecure_port = None
            for env_ent in i.spec.template.spec.containers[0].env:
                if env_ent.name == "REKCURD_GRPC_PROTO_VERSION":
                    version = env_ent.value
                elif env_ent.name == "REKCURD_MODEL_FILE_PATH":
                    filepath = env_ent.value
                elif env_ent.name == "REKCURD_SERVICE_INSECURE_HOST":
                    insecure_host = env_ent.value
                elif env_ent.name == "REKCURD_SERVICE_INSECURE_PORT":
                    insecure_port = int(env_ent.value)

            """Model registration."""
            model_model: ModelModel = db.session.query(ModelModel).filter(
                ModelModel.application_id == application_id,
                ModelModel.filepath == filepath).one_or_none()
            if model_model is None:
                model_model = ModelModel(application_id=application_id,
                                         filepath=filepath, description="Automatically registered.")
                db.session.add(model_model)
                db.session.flush()

            """Service registration."""
            service_model = ServiceModel(
                service_id=service_id,
                application_id=application_id,
                display_name="{}-{}".format(service_level, service_id),
                service_level=service_level,
                version=version,
                model_id=model_model.model_id,
                insecure_host=insecure_host,
                insecure_port=insecure_port)
            db.session.add(service_model)
            db.session.flush()
    return


def apply_rekcurd_to_kubernetes(
        project_id: int, application_id: str, service_level: str, version: str,
        insecure_host: str, insecure_port: int, replicas_default: int, replicas_minimum: int,
        replicas_maximum: int, autoscale_cpu_threshold: str, policy_max_surge: int,
        policy_max_unavailable: int, policy_wait_seconds: int, container_image: str,
        resource_request_cpu: str, resource_request_memory: str, resource_limit_cpu: str,
        resource_limit_memory: str, commit_message: str, service_model_assignment: int,
        service_git_url: str = "", service_git_branch: str = "", service_boot_script: str = "",
        debug_mode: bool = False, service_id: str = None,
        display_name: str = None, description: str = None,
        **kwargs) -> str:
    """
    kubectl apply
    :param project_id:
    :param application_id:
    :param service_level:
    :param version:
    :param insecure_host:
    :param insecure_port:
    :param replicas_default:
    :param replicas_minimum:
    :param replicas_maximum:
    :param autoscale_cpu_threshold:
    :param policy_max_surge:
    :param policy_max_unavailable:
    :param policy_wait_seconds:
    :param container_image:
    :param resource_request_cpu:
    :param resource_request_memory:
    :param resource_limit_cpu:
    :param resource_limit_memory:
    :param commit_message:
    :param service_model_assignment:
    :param service_git_url:
    :param service_git_branch:
    :param service_boot_script:
    :param debug_mode:
    :param service_id:
    :param display_name:
    :param description:
    :param kwargs:
    :return:
    """
    __num_retry = 5
    progress_deadline_seconds = \
        int(__num_retry*policy_wait_seconds*replicas_maximum/(policy_max_surge+policy_max_unavailable))
    is_creation_mode = False
    if service_id is None:
        is_creation_mode = True
        service_id = uuid.uuid4().hex
    data_server_model: DataServerModel = db.session.query(DataServerModel).filter(
        DataServerModel.project_id == project_id).first_or_404()
    application_model: ApplicationModel = db.session.query(ApplicationModel).filter(
        ApplicationModel.application_id == application_id).first_or_404()
    application_name = application_model.application_name
    model_model: ModelModel = db.session.query(ModelModel).filter(
        ModelModel.model_id == service_model_assignment).first_or_404()

    for kubernetes_model in db.session.query(KubernetesModel).filter(KubernetesModel.project_id == project_id).all():
        full_config_path = get_full_config_path(kubernetes_model.config_path)
        from kubernetes import client, config
        config.load_kube_config(full_config_path)

        pod_env = [
            client.V1EnvVar(
                name="REKCURD_SERVICE_UPDATE_FLAG",
                value=commit_message
            ),
            client.V1EnvVar(
                name="REKCURD_KUBERNETES_MODE",
                value="True"
            ),
            client.V1EnvVar(
                name="REKCURD_DEBUG_MODE",
                value=str(debug_mode)
            ),
            client.V1EnvVar(
                name="REKCURD_APPLICATION_NAME",
                value=application_name
            ),
            client.V1EnvVar(
                name="REKCURD_SERVICE_INSECURE_HOST",
                value=insecure_host
            ),
            client.V1EnvVar(
                name="REKCURD_SERVICE_INSECURE_PORT",
                value=str(insecure_port)
            ),
            client.V1EnvVar(
                name="REKCURD_SERVICE_ID",
                value=service_id
            ),
            client.V1EnvVar(
                name="REKCURD_SERVICE_LEVEL",
                value=service_level
            ),
            client.V1EnvVar(
                name="REKCURD_GRPC_PROTO_VERSION",
                value=version
            ),
            client.V1EnvVar(
                name="REKCURD_MODEL_MODE",
                value=data_server_model.data_server_mode.value
            ),
            client.V1EnvVar(
                name="REKCURD_MODEL_FILE_PATH",
                value=model_model.filepath
            ),
            client.V1EnvVar(
                name="REKCURD_CEPH_ACCESS_KEY",
                value=str(data_server_model.ceph_access_key or "xxx")
            ),
            client.V1EnvVar(
                name="REKCURD_CEPH_SECRET_KEY",
                value=str(data_server_model.ceph_secret_key or "xxx")
            ),
            client.V1EnvVar(
                name="REKCURD_CEPH_HOST",
                value=str(data_server_model.ceph_host or "xxx")
            ),
            client.V1EnvVar(
                name="REKCURD_CEPH_PORT",
                value=str(data_server_model.ceph_port or "1234")
            ),
            client.V1EnvVar(
                name="REKCURD_CEPH_IS_SECURE",
                value=str(data_server_model.ceph_is_secure or "False")
            ),
            client.V1EnvVar(
                name="REKCURD_CEPH_BUCKET_NAME",
                value=str(data_server_model.ceph_bucket_name or "xxx")
            ),
            client.V1EnvVar(
                name="REKCURD_AWS_ACCESS_KEY",
                value=str(data_server_model.aws_access_key or "xxx")
            ),
            client.V1EnvVar(
                name="REKCURD_AWS_SECRET_KEY",
                value=str(data_server_model.aws_secret_key or "xxx")
            ),
            client.V1EnvVar(
                name="REKCURD_AWS_BUCKET_NAME",
                value=str(data_server_model.aws_bucket_name or "xxx")
            ),
            # TODO: GCP
            client.V1EnvVar(
                name="REKCURD_SERVICE_GIT_URL",
                value=service_git_url
            ),
            client.V1EnvVar(
                name="REKCURD_SERVICE_GIT_BRANCH",
                value=service_git_branch
            ),
            client.V1EnvVar(
                name="REKCURD_SERVICE_BOOT_SHELL",
                value=service_boot_script
            ),
        ]

        """Namespace registration."""
        core_vi_api = client.CoreV1Api()
        try:
            core_vi_api.read_namespace(name=service_level)
        except:
            api.logger.info("\"{}\" namespace created".format(service_level))
            v1_namespace = client.V1Namespace(
                api_version="v1",
                kind="Namespace",
                metadata=client.V1ObjectMeta(
                    name=service_level
                )
            )
            core_vi_api.create_namespace(
                body=v1_namespace
            )

        """Create/patch Deployment."""
        v1_deployment = client.V1Deployment(
            api_version="apps/v1",
            kind="Deployment",
            metadata=client.V1ObjectMeta(
                name="deploy-{0}".format(service_id),
                namespace=service_level,
                labels={"rekcurd-worker": "True", "id": application_id,
                        "name": application_name, "sel": service_id}
            ),
            spec=client.V1DeploymentSpec(
                min_ready_seconds=policy_wait_seconds,
                progress_deadline_seconds=progress_deadline_seconds,
                replicas=replicas_default,
                revision_history_limit=3,
                selector=client.V1LabelSelector(
                    match_labels={"sel": service_id}
                ),
                strategy=client.V1DeploymentStrategy(
                    type="RollingUpdate",
                    rolling_update=client.V1RollingUpdateDeployment(
                        max_surge=policy_max_surge,
                        max_unavailable=policy_max_unavailable)
                ),
                template=client.V1PodTemplateSpec(
                    metadata=client.V1ObjectMeta(
                        labels={"rekcurd-worker": "True", "id": application_id,
                                "name": application_name, "sel": service_id}
                    ),
                    spec=client.V1PodSpec(
                        affinity=client.V1Affinity(
                            pod_anti_affinity=client.V1PodAntiAffinity(
                                preferred_during_scheduling_ignored_during_execution=[
                                    client.V1WeightedPodAffinityTerm(
                                        pod_affinity_term=client.V1PodAffinityTerm(
                                            label_selector=client.V1LabelSelector(
                                                match_expressions=[
                                                    client.V1LabelSelectorRequirement(
                                                        key="id",
                                                        operator="In",
                                                        values=[service_id]
                                                    )
                                                ]
                                            ),
                                            topology_key="kubernetes.io/hostname"
                                        ),
                                        weight=100
                                    )
                                ]
                            )
                        ),
                        containers=[
                            client.V1Container(
                                env=pod_env,
                                image=container_image,
                                image_pull_policy="Always",
                                name=service_id,
                                ports=[
                                    client.V1ContainerPort(container_port=insecure_port)
                                ],
                                resources=client.V1ResourceRequirements(
                                    limits={
                                        "cpu": str(resource_limit_cpu),
                                        "memory": resource_limit_memory
                                    },
                                    requests={
                                        "cpu": str(resource_request_cpu),
                                        "memory": resource_request_memory
                                    }
                                ),
                                security_context=client.V1SecurityContext(
                                    privileged=True
                                )
                            )
                        ],
                        node_selector={"host": service_level}
                    )
                )
            )
        )
        apps_v1_api = client.AppsV1Api()
        if is_creation_mode:
            api.logger.info("Deployment created.")
            apps_v1_api.create_namespaced_deployment(
                body=v1_deployment,
                namespace=service_level
            )
        else:
            api.logger.info("Deployment patched.")
            apps_v1_api.patch_namespaced_deployment(
                body=v1_deployment,
                name="deploy-{0}".format(service_id),
                namespace=service_level
            )

        """Create/patch Service."""
        v1_service = client.V1Service(
            api_version="v1",
            kind="Service",
            metadata=client.V1ObjectMeta(
                name="svc-{0}".format(service_id),
                namespace=service_level,
                labels={"rekcurd-worker": "True", "id": application_id,
                        "name": application_name, "sel": service_id}
            ),
            spec=client.V1ServiceSpec(
                ports=[
                    client.V1ServicePort(
                        name="grpc-backend",
                        port=insecure_port,
                        protocol="TCP",
                        target_port=insecure_port
                    )
                ],
                selector={"sel": service_id}
            )
        )
        core_vi_api = client.CoreV1Api()
        if is_creation_mode:
            api.logger.info("Service created.")
            core_vi_api.create_namespaced_service(
                namespace=service_level,
                body=v1_service
            )
        else:
            api.logger.info("Service patched.")
            core_vi_api.patch_namespaced_service(
                namespace=service_level,
                name="svc-{0}".format(service_id),
                body=v1_service
            )

        """Create/patch Autoscaler."""
        v1_horizontal_pod_autoscaler = client.V1HorizontalPodAutoscaler(
            api_version="autoscaling/v1",
            kind="HorizontalPodAutoscaler",
            metadata=client.V1ObjectMeta(
                name="hpa-{0}".format(service_id),
                namespace=service_level,
                labels={"rekcurd-worker": "True", "id": application_id,
                        "name": application_name, "sel": service_id}
            ),
            spec=client.V1HorizontalPodAutoscalerSpec(
                max_replicas=replicas_maximum,
                min_replicas=replicas_minimum,
                scale_target_ref=client.V1CrossVersionObjectReference(
                    api_version="apps/v1",
                    kind="Deployment",
                    name="deploy-{0}".format(service_id)
                ),
                target_cpu_utilization_percentage=autoscale_cpu_threshold
            )
        )
        autoscaling_v1_api = client.AutoscalingV1Api()
        if is_creation_mode:
            api.logger.info("Autoscaler created.")
            autoscaling_v1_api.create_namespaced_horizontal_pod_autoscaler(
                namespace=service_level,
                body=v1_horizontal_pod_autoscaler
            )
        else:
            api.logger.info("Autoscaler patched.")
            autoscaling_v1_api.patch_namespaced_horizontal_pod_autoscaler(
                namespace=service_level,
                name="hpa-{0}".format(service_id),
                body=v1_horizontal_pod_autoscaler
            )

        """Create Istio ingress if this is the first application."""
        custom_object_api = client.CustomObjectsApi()
        try:
            custom_object_api.get_namespaced_custom_object(
                group="networking.istio.io",
                version="v1alpha3",
                namespace=service_level,
                plural="virtualservices",
                name="ing-vs-{0}".format(application_id),
            )
        except:
            ingress_virtual_service_body = {
                "apiVersion": "networking.istio.io/v1alpha3",
                "kind": "VirtualService",
                "metadata": {
                    "labels": {"rekcurd-worker": "True", "id": application_id,
                               "name": application_name, "sel": service_id},
                    "name": "ing-vs-{0}".format(application_id),
                    "namespace": service_level
                },
                "spec": {
                    "hosts": [
                        "*"
                    ],
                    "gateways": [
                        "rekcurd-ingress-gateway"
                    ],
                    "http": [
                        {
                            "match": [
                                {
                                    "headers": {
                                        "x-rekcurd-application-name": {
                                            "exact": application_name
                                        },
                                        "x-rekcurd-sevice-level": {
                                            "exact": service_level
                                        },
                                        "x-rekcurd-grpc-version": {
                                            "exact": version
                                        },
                                    }
                                }
                            ],
                            "route": [
                                {
                                    "destination": {
                                        "port": {
                                            "number": insecure_port
                                        },
                                        "host": "svc-{0}".format(service_id)
                                    },
                                    "weight": 100
                                }
                            ],
                            "retries": {
                                "attempts": 25,
                                "perTryTimeout": "1s"
                            }
                        }
                    ]
                }
            }
            api.logger.info("Istio created.")
            custom_object_api.create_namespaced_custom_object(
                group="networking.istio.io",
                version="v1alpha3",
                namespace=service_level,
                plural="virtualservices",
                body=ingress_virtual_service_body
            )

        """Add service model."""
        if is_creation_mode:
            if display_name is None:
                display_name = "{0}-{1}".format(service_level, service_id)
            service_model = ServiceModel(
                service_id=service_id, application_id=application_id, display_name=display_name,
                description=description, service_level=service_level, version=version,
                model_id=service_model_assignment, insecure_host=insecure_host,
                insecure_port=insecure_port)
            db.session.add(service_model)
            db.session.flush()

    """Finish."""
    return service_id


def delete_kubernetes_deployment(kubernetes_models: list, application_id: str, service_id: str):
    """
    Delete Kubernetes deployment.
    :param kubernetes_models:
    :param application_id:
    :param service_id:
    :return:
    """
    service_model: ServiceModel = db.session.query(ServiceModel).filter(ServiceModel.service_id == service_id).first_or_404()
    for kubernetes_model in kubernetes_models:
        full_config_path = get_full_config_path(kubernetes_model.config_path)
        from kubernetes import client, config
        config.load_kube_config(full_config_path)
        """Deployment"""
        apps_v1_api = client.AppsV1Api()
        apps_v1_api.delete_namespaced_deployment(
            name="deploy-{0}".format(service_id),
            namespace=service_model.service_level,
            body=client.V1DeleteOptions()
        )
        """Service"""
        core_vi_api = client.CoreV1Api()
        core_vi_api.delete_namespaced_service(
            name="svc-{0}".format(service_id),
            namespace=service_model.service_level,
            body=client.V1DeleteOptions()
        )
        """Autoscaler"""
        autoscaling_v1_api = client.AutoscalingV1Api()
        autoscaling_v1_api.delete_namespaced_horizontal_pod_autoscaler(
            name="hpa-{0}".format(service_id),
            namespace=service_model.service_level,
            body=client.V1DeleteOptions()
        )
        """Istio"""
        custom_object_api = client.CustomObjectsApi()
        ingress_virtual_service_body = custom_object_api.get_namespaced_custom_object(
            group="networking.istio.io",
            version="v1alpha3",
            namespace=service_model.service_level,
            plural="virtualservices",
            name="ing-vs-{0}".format(application_id),
        )
        routes = ingress_virtual_service_body["spec"]["http"][0]["route"]
        new_routes = [route for route in routes if route["destination"]["host"] != "svc-{0}".format(service_id)]
        if new_routes:
            weights = [route["weight"] for route in new_routes]
            norm_factor = 100.0/sum(weights)
            for route in new_routes:
                route["weight"] = round(route["weight"]*norm_factor)
            ingress_virtual_service_body["spec"]["http"][0]["route"] = new_routes
            custom_object_api.patch_namespaced_custom_object(
                group="networking.istio.io",
                version="v1alpha3",
                namespace=service_model.service_level,
                plural="virtualservices",
                name="ing-vs-{0}".format(application_id),
                body=ingress_virtual_service_body
            )
        else:
            custom_object_api.delete_namespaced_custom_object(
                group="networking.istio.io",
                version="v1alpha3",
                namespace=service_model.service_level,
                plural="virtualservices",
                name="ing-vs-{0}".format(application_id),
                body=client.V1DeleteOptions()
            )
    db.session.query(ServiceModel).filter(ServiceModel.service_id == service_id).delete()
    db.session.flush()
    return


def apply_new_route_weight(
        project_id: int, application_id: str, service_level: str, service_ids: list, service_weights: list):
    service_weight_dict = dict()
    service_weight_checker = list()
    for i in range(len(service_ids)):
        key = service_ids[i]
        service_weight_dict[key] = int(service_weights[i])
        service_weight_checker.append(int(service_weights[i]))
    if sum(service_weight_checker) != 100:
        raise RekcurdDashboardException("total weight must be 100.")
    for kubernetes_model in db.session.query(KubernetesModel).filter(KubernetesModel.project_id == project_id).all():
        full_config_path = get_full_config_path(kubernetes_model.config_path)
        from kubernetes import client, config
        config.load_kube_config(full_config_path)
        custom_object_api = client.CustomObjectsApi()
        ingress_virtual_service_body = custom_object_api.get_namespaced_custom_object(
            group="networking.istio.io",
            version="v1alpha3",
            namespace=service_level,
            plural="virtualservices",
            name="ing-vs-{0}".format(application_id),
        )
        routes = []
        for service_id in service_ids:
            service_model: ServiceModel = db.session.query(ServiceModel).filter(
                ServiceModel.service_id == service_id).first_or_404()
            route = {
                "destination": {
                    "port": {
                        "number": service_model.insecure_port
                    },
                    "host": "svc-{0}".format(service_id)
                },
                "weight": service_weight_dict[service_id]
            }
            routes.append(route)
        ingress_virtual_service_body["spec"]["http"][0]["route"] = routes
        custom_object_api.patch_namespaced_custom_object(
            group="networking.istio.io",
            version="v1alpha3",
            namespace=service_level,
            plural="virtualservices",
            name="ing-vs-{0}".format(application_id),
            body=ingress_virtual_service_body
        )
    return


def load_kubernetes_deployment_info(project_id: int, application_id: str, service_id: str) -> dict:
    """
    Load deployment info from Kubernetes.
    :param project_id:
    :param application_id:
    :param service_id:
    :return:
    """
    kubernetes_model: KubernetesModel = db.session.query(KubernetesModel).filter(
        KubernetesModel.project_id == project_id).first()
    service_model: ServiceModel = db.session.query(ServiceModel).filter(
        ServiceModel.service_id == service_id).first_or_404()

    full_config_path = get_full_config_path(kubernetes_model.config_path)
    from kubernetes import client, config
    config.load_kube_config(full_config_path)

    apps_v1_api = client.AppsV1Api()
    v1_deployment = apps_v1_api.read_namespaced_deployment(
        name="deploy-{0}".format(service_id),
        namespace=service_model.service_level
    )
    autoscaling_v1_api = client.AutoscalingV1Api()
    v1_horizontal_pod_autoscaler = autoscaling_v1_api.read_namespaced_horizontal_pod_autoscaler(
        name="hpa-{0}".format(service_id),
        namespace=service_model.service_level
    )

    deployment_info = {}
    filepath = None
    deployment_info["application_name"] = v1_deployment.metadata.labels["name"]
    deployment_info["service_id"] = service_id
    for env_ent in v1_deployment.spec.template.spec.containers[0].env:
        if env_ent.name == "REKCURD_SERVICE_UPDATE_FLAG":
            deployment_info["commit_message"] = env_ent.value
        elif env_ent.name == "REKCURD_SERVICE_INSECURE_HOST":
            deployment_info["insecure_host"] = env_ent.value
        elif env_ent.name == "REKCURD_SERVICE_INSECURE_PORT":
            deployment_info["insecure_port"] = int(env_ent.value)
        elif env_ent.name == "REKCURD_SERVICE_LEVEL":
            deployment_info["service_level"] = env_ent.value
        elif env_ent.name == "REKCURD_GRPC_PROTO_VERSION":
            deployment_info["version"] = env_ent.value
        elif env_ent.name == "REKCURD_MODEL_FILE_PATH":
            filepath = env_ent.value
        elif env_ent.name == "REKCURD_SERVICE_GIT_URL":
            deployment_info["service_git_url"] = env_ent.value
        elif env_ent.name == "REKCURD_SERVICE_GIT_BRANCH":
            deployment_info["service_git_branch"] = env_ent.value
        elif env_ent.name == "REKCURD_SERVICE_BOOT_SHELL":
            deployment_info["service_boot_script"] = env_ent.value
    model_model: ModelModel = db.session.query(ModelModel).filter(
        ModelModel.application_id == application_id, ModelModel.filepath == filepath).first_or_404()
    deployment_info["service_model_assignment"] = model_model.model_id
    deployment_info["replicas_default"] = v1_deployment.spec.replicas
    deployment_info["replicas_minimum"] = v1_horizontal_pod_autoscaler.spec.min_replicas
    deployment_info["replicas_maximum"] = v1_horizontal_pod_autoscaler.spec.max_replicas
    deployment_info["autoscale_cpu_threshold"] = v1_horizontal_pod_autoscaler.spec.target_cpu_utilization_percentage
    deployment_info["policy_max_surge"] = v1_deployment.spec.strategy.rolling_update.max_surge
    deployment_info["policy_max_unavailable"] = v1_deployment.spec.strategy.rolling_update.max_unavailable
    deployment_info["policy_wait_seconds"] = v1_deployment.spec.min_ready_seconds
    deployment_info["container_image"] = v1_deployment.spec.template.spec.containers[0].image
    deployment_info["resource_request_cpu"] = \
        kubernetes_cpu_to_float(v1_deployment.spec.template.spec.containers[0].resources.requests["cpu"])
    deployment_info["resource_request_memory"] = \
        v1_deployment.spec.template.spec.containers[0].resources.requests["memory"]
    deployment_info["resource_limit_cpu"] = \
        kubernetes_cpu_to_float(v1_deployment.spec.template.spec.containers[0].resources.limits["cpu"])
    deployment_info["resource_limit_memory"] = \
        v1_deployment.spec.template.spec.containers[0].resources.limits["memory"]
    return deployment_info


def switch_model_assignment(project_id: int, application_id: str, service_id: str, model_id: int):
    """
    Switch model assignment.
    :param project_id:
    :param application_id:
    :param service_id:
    :param model_id:
    :return:
    """
    service_model: ServiceModel = db.session.query(ServiceModel).filter(
        ServiceModel.service_id == service_id).first_or_404()
    deployment_info = load_kubernetes_deployment_info(project_id, application_id, service_id)
    deployment_info["service_model_assignment"] = model_id
    deployment_info["commit_message"] = "model_id={0} on {1:%Y%m%d%H%M%S}".format(model_id, datetime.utcnow())

    apply_rekcurd_to_kubernetes(
        project_id=project_id, application_id=application_id, **deployment_info)

    service_model.model_id = model_id
    db.session.flush()
    return


def backup_kubernetes_deployment(
        kubernetes_model: KubernetesModel, application_model: ApplicationModel, service_model: ServiceModel):
    """
    Backup Kubernetes deployment.
    :param kubernetes_model:
    :param application_model:
    :param service_model:
    :return:
    """
    full_config_path = get_full_config_path(kubernetes_model.config_path)
    from kubernetes import client, config
    config.load_kube_config(full_config_path)
    save_dir = Path(api.dashboard_config.DIR_KUBE_CONFIG, application_model.application_name)
    save_dir.mkdir(parents=True, exist_ok=True)
    api_client = client.ApiClient()

    """Deployment"""
    apps_v1_api = client.AppsV1Api()
    v1_deployment = apps_v1_api.read_namespaced_deployment(
        name="deploy-{0}".format(service_model.service_id),
        namespace=service_model.service_level,
        exact=True,
        export=True
    )
    json.dump(api_client.sanitize_for_serialization(v1_deployment),
              Path(save_dir, "deploy-{0}.json".format(service_model.service_id)).open("w", encoding='utf-8'),
              ensure_ascii=False, indent=2)
    """Service"""
    core_vi_api = client.CoreV1Api()
    v1_service = core_vi_api.read_namespaced_service(
        name="svc-{0}".format(service_model.service_id),
        namespace=service_model.service_level,
        exact=True,
        export=True
    )
    json.dump(api_client.sanitize_for_serialization(v1_service),
              Path(save_dir, "svc-{0}.json".format(service_model.service_id)).open("w", encoding='utf-8'),
              ensure_ascii=False, indent=2)
    """Autoscaler"""
    autoscaling_v1_api = client.AutoscalingV1Api()
    v1_horizontal_pod_autoscaler = autoscaling_v1_api.read_namespaced_horizontal_pod_autoscaler(
        name="hpa-{0}".format(service_model.service_id),
        namespace=service_model.service_level,
        exact=True,
        export=True
    )
    json.dump(api_client.sanitize_for_serialization(v1_horizontal_pod_autoscaler),
              Path(save_dir, "hpa-{0}.json".format(service_model.service_id)).open("w", encoding='utf-8'),
              ensure_ascii=False, indent=2)
    return


def load_istio_routing(kubernetes_model: KubernetesModel, application_model: ApplicationModel, service_level: str):
    """
    Load istio routing info from Kubernetes.
    :param kubernetes_model:
    :param application_model:
    :param service_level:
    :return:
    """
    full_config_path = get_full_config_path(kubernetes_model.config_path)
    from kubernetes import client, config
    config.load_kube_config(full_config_path)
    custom_object_api = client.CustomObjectsApi()
    ingress_virtual_service_body = custom_object_api.get_namespaced_custom_object(
        group="networking.istio.io",
        version="v1alpha3",
        namespace=service_level,
        plural="virtualservices",
        name="ing-vs-{0}".format(application_model.application_id),
    )
    return ingress_virtual_service_body["spec"]["http"][0]["route"]


def backup_istio_routing(kubernetes_model: KubernetesModel, application_model: ApplicationModel, service_level: str):
    """
    Backup Kubernetes deployment.
    :param kubernetes_model:
    :param application_model:
    :param service_level:
    :return:
    """
    full_config_path = get_full_config_path(kubernetes_model.config_path)
    from kubernetes import client, config
    config.load_kube_config(full_config_path)
    save_dir = Path(api.dashboard_config.DIR_KUBE_CONFIG, application_model.application_name)
    save_dir.mkdir(parents=True, exist_ok=True)

    """Istio"""
    custom_object_api = client.CustomObjectsApi()
    ingress_virtual_service_body = custom_object_api.get_namespaced_custom_object(
        group="networking.istio.io",
        version="v1alpha3",
        namespace=service_level,
        plural="virtualservices",
        name="ing-vs-{0}".format(application_model.application_id),
    )
    json.dump(ingress_virtual_service_body,
              Path(save_dir, "ing-vs-{0}.json".format(
                  application_model.application_id)).open("w", encoding='utf-8'), ensure_ascii=False, indent=2)
    return



# coding: utf-8


import traceback

from flask_restplus import Api
from flask_restplus.utils import default_id
from werkzeug.exceptions import HTTPException
from kubernetes.client.rest import ApiException

from rekcurd_dashboard.utils import RekcurdDashboardException, ProjectUserRoleException, ApplicationUserRoleException
from rekcurd_dashboard.auth import auth_required
from rekcurd_dashboard.models import db


class RekcurdDashboardApi(Api):
    def __init__(self, app=None, version='1.0', title=None, description=None,
                 terms_url=None, license=None, license_url=None,
                 contact=None, contact_url=None, contact_email=None,
                 authorizations=None, security=None, doc='/', default_id=default_id,
                 default='default', default_label='Default namespace', validate=None,
                 tags=None, prefix='', ordered=False,
                 default_mediatype='application/json', decorators=None,
                 catch_all_404s=False, serve_challenge_on_401=False, format_checker=None,
                 **kwargs):
        super().__init__(app=app, version=version, title=title, description=description,
                         terms_url=terms_url, license=license, license_url=license_url,
                         contact=contact, contact_url=contact_url, contact_email=contact_email,
                         authorizations=authorizations, security=security, doc=doc, default_id=default_id,
                         default=default, default_label=default_label, validate=validate,
                         tags=tags, prefix=prefix, ordered=ordered,
                         default_mediatype=default_mediatype, decorators=decorators,
                         catch_all_404s=catch_all_404s, serve_challenge_on_401=serve_challenge_on_401,
                         format_checker=format_checker,
                         **kwargs)
        self.dashboard_config = None
        self.logger = None

    def init_app(self, app, **kwargs):
        super().init_app(app, **kwargs)
        self.dashboard_config = kwargs.get('dashboard_config')
        self.logger = kwargs.get('logger')


api = RekcurdDashboardApi(
    version='1.0',
    title='Rekcurd dashboard API',
    description='Rekcurd dashboard API',
    doc='/doc/',
    decorators=[auth_required]
)


from .common import DatetimeToTimestamp, kubernetes_cpu_to_float, status_model
from .kubernetes_handler import (
    get_full_config_path, save_kubernetes_access_file, remove_kubernetes_access_file,
    update_kubernetes_deployment_info, apply_rekcurd_to_kubernetes, load_kubernetes_deployment_info,
    switch_model_assignment, backup_kubernetes_deployment, delete_kubernetes_deployment,
    backup_istio_routing, load_istio_routing, apply_new_route_weight
)
from .api_admin import admin_api_namespace
from .api_project import project_api_namespace
from .api_data_server import data_server_api_namespace
from .api_kubernetes import kubernetes_api_namespace
from .api_application import application_api_namespace
from .api_service import service_api_namespace
from .api_model import model_api_namespace
from .api_service_deployment import service_deployment_api_namespace
from .api_service_routing import service_routing_api_namespace
from .api_evaluation import evaluation_api_namespace
from .api_misc import misc_api_namespace


@api.errorhandler(ApiException)
def api_exception_handler(error):
    api.logger.error(str(error))
    api.logger.error(traceback.format_exc())
    return {'status': False, 'message': 'Something wrong with accessing Kubernetes'}, 400


@api.errorhandler(RekcurdDashboardException)
@api.errorhandler(ProjectUserRoleException)
@api.errorhandler(ApplicationUserRoleException)
def rekcurd_exception_handler(error):
    api.logger.error(str(error))
    api.logger.error(traceback.format_exc())
    db.session.rollback()
    db.session.close()
    return {'status': False, 'message': str(error)}, 400


@api.errorhandler
def default_error_handler(error):
    api.logger.error(str(error))
    api.logger.error(traceback.format_exc())
    db.session.rollback()
    db.session.close()
    return {'status': False, 'message': 'Something wrong. Contact the admin.'}, 500


api.add_namespace(admin_api_namespace, path='/api')
api.add_namespace(project_api_namespace, path='/api')
api.add_namespace(data_server_api_namespace, path='/api')
api.add_namespace(kubernetes_api_namespace, path='/api')
api.add_namespace(application_api_namespace, path='/api')
api.add_namespace(service_api_namespace, path='/api')
api.add_namespace(model_api_namespace, path='/api')
api.add_namespace(service_deployment_api_namespace, path='/api')
api.add_namespace(service_routing_api_namespace, path='/api')
api.add_namespace(evaluation_api_namespace, path='/api')
api.add_namespace(misc_api_namespace, path='/api')



# coding: utf-8


from flask_restplus import Resource, Namespace

from . import api


misc_api_namespace = Namespace('misc', description='Misc API Endpoint.')


@misc_api_namespace.route('/settings')
class ApiSettings(Resource):
    def get(self):
        result = {
            'auth': api.dashboard_config.IS_ACTIVATE_AUTH
        }
        return result



# -*- coding: utf-8 -*-


from rekcurd_dashboard.core import create_app


def server_handler(args: dict):
    app, config = create_app(config_file=args["settings"], logger_file=args["logger"], **args)
    app.run(host=args["host"], port=config.SERVICE_PORT, threaded=True)



from flask_jwt_simple import get_jwt_identity
from flask_restplus import Resource, Namespace, fields, reqparse

from rekcurd_dashboard.models import db, UserModel, ApplicationRole, ProjectUserRoleModel, ApplicationUserRoleModel
from rekcurd_dashboard.utils import RekcurdDashboardException
from . import status_model


admin_api_namespace = Namespace('admin', description='Admin API Endpoint.')
success_or_not = admin_api_namespace.model('Status', status_model)
user_info = admin_api_namespace.model('User', {
    'auth_id': fields.String(required=True),
    'user_name': fields.String(required=True)
})
project_role_info = admin_api_namespace.model('Role', {
    'project_role': fields.String(required=True),
    'user': fields.Nested(user_info, required=True)
})
application_role_info = admin_api_namespace.model('Role', {
    'application_role': fields.String(required=True),
    'user': fields.Nested(user_info, required=True)
})


@admin_api_namespace.route('/users')
class ApiUsers(Resource):
    @admin_api_namespace.marshal_list_with(user_info)
    def get(self):
        return UserModel.query.all()


@admin_api_namespace.route('/projects/<int:project_id>/acl')
class ApiProjectIdACL(Resource):
    save_acl_parser = reqparse.RequestParser()
    save_acl_parser.add_argument('uid', type=str, required=True, location='form')
    save_acl_parser.add_argument('role', type=str, required=True, location='form',
                                 choices=('admin', 'member'))

    @admin_api_namespace.marshal_list_with(project_role_info)
    def get(self, project_id):
        roles = ProjectUserRoleModel.query.filter_by(project_id=project_id).all()
        return roles

    @admin_api_namespace.marshal_with(success_or_not)
    @admin_api_namespace.expect(save_acl_parser)
    def post(self, project_id):
        args = self.save_acl_parser.parse_args()
        uid = args['uid']
        user_model = db.session.query(UserModel).filter(UserModel.auth_id == uid).one_or_none()
        if user_model is None:
            raise RekcurdDashboardException("No user found.")

        role = args['role']
        project_user_role_model = db.session.query(ProjectUserRoleModel).filter(
            ProjectUserRoleModel.project_id == project_id,
            ProjectUserRoleModel.user_id == user_model.user_id).one_or_none()
        if project_user_role_model is None:
            project_user_role_model = ProjectUserRoleModel(
                project_id=project_id,
                user_id=user_model.user_id,
                project_role=role)
            db.session.add(project_user_role_model)
            db.session.commit()
            db.session.close()
            return {"status": True, "message": "Success."}, 200
        else:
            raise RekcurdDashboardException("Already assigned.")

    @admin_api_namespace.marshal_with(success_or_not)
    @admin_api_namespace.expect(save_acl_parser)
    def patch(self, project_id):
        args = self.save_acl_parser.parse_args()
        uid = args['uid']
        user_model = db.session.query(UserModel).filter(UserModel.auth_id == uid).one_or_none()
        if user_model is None:
            raise RekcurdDashboardException("No user found.")
        project_user_role_model = db.session.query(ProjectUserRoleModel).filter(
            ProjectUserRoleModel.project_id == project_id,
            ProjectUserRoleModel.user_id == user_model.user_id).first_or_404()
        project_user_role_model.project_role = args['role']
        db.session.commit()
        db.session.close()
        return {"status": True, "message": "Success."}, 200


@admin_api_namespace.route('/projects/<int:project_id>/acl/users/<uid>')
class ApiProjectIdUserIdACL(Resource):
    @admin_api_namespace.marshal_with(success_or_not)
    def delete(self, project_id, uid):
        user_model = db.session.query(UserModel).filter(UserModel.auth_id == uid).one_or_none()
        if user_model is None:
            raise RekcurdDashboardException("No user found.")

        ProjectUserRoleModel.query.filter(
            ProjectUserRoleModel.project_id == project_id,
            ProjectUserRoleModel.user_id == user_model.user_id).delete()
        db.session.commit()
        db.session.close()
        return {"status": True, "message": "Success."}, 200


@admin_api_namespace.route('/projects/<int:project_id>/applications/<application_id>/acl')
class ApiApplicationIdACL(Resource):
    save_acl_parser = reqparse.RequestParser()
    save_acl_parser.add_argument('uid', type=str, required=True, location='form')
    save_acl_parser.add_argument('role', type=str, required=True, location='form',
                                 choices=('admin', 'editor', 'viewer'))

    @admin_api_namespace.marshal_list_with(application_role_info)
    def get(self, project_id, application_id):
        roles = ApplicationUserRoleModel.query.filter_by(application_id=application_id).all()
        return roles

    @admin_api_namespace.marshal_with(success_or_not)
    @admin_api_namespace.expect(save_acl_parser)
    def post(self, project_id, application_id):
        args = self.save_acl_parser.parse_args()
        uid = args['uid']
        user_model = db.session.query(UserModel).filter(UserModel.auth_id == uid).one_or_none()
        if user_model is None:
            raise RekcurdDashboardException("No user found.")

        role = args['role']
        # if role is added by first user, they will be owner
        roles = db.session.query(ApplicationUserRoleModel).filter(
            ApplicationUserRoleModel.application_id == application_id).count()
        if roles == 0:
            sender_id = get_jwt_identity()
            if user_model.user_id != sender_id:
                ownerObj = ApplicationUserRoleModel(
                    application_id=application_id,
                    user_id=sender_id,
                    application_role=ApplicationRole.admin.name)
                db.session.add(ownerObj)
                db.session.flush()
            else:
                role = ApplicationRole.admin.name
        application_user_role_model = db.session.query(ApplicationUserRoleModel).filter(
            ApplicationUserRoleModel.application_id == application_id,
            ApplicationUserRoleModel.user_id == user_model.user_id).one_or_none()
        if application_user_role_model is None:
            application_user_role_model = ApplicationUserRoleModel(
                application_id=application_id,
                user_id=user_model.user_id,
                application_role=role)
            db.session.add(application_user_role_model)
            db.session.commit()
            db.session.close()
            return {"status": True, "message": "Success."}, 200
        else:
            raise RekcurdDashboardException("Already assigned.")

    @admin_api_namespace.marshal_with(success_or_not)
    @admin_api_namespace.expect(save_acl_parser)
    def patch(self, project_id, application_id):
        args = self.save_acl_parser.parse_args()
        uid = args['uid']
        user_model = db.session.query(UserModel).filter(UserModel.auth_id == uid).one_or_none()
        if user_model is None:
            raise RekcurdDashboardException("No user found.")
        application_user_role_model: ApplicationUserRoleModel = db.session.query(ApplicationUserRoleModel).filter(
            ApplicationUserRoleModel.application_id == application_id,
            ApplicationUserRoleModel.user_id == user_model.user_id).first_or_404()
        application_user_role_model.application_role = args['role']
        db.session.commit()
        db.session.close()
        return {"status": True, "message": "Success."}, 200


@admin_api_namespace.route('/projects/<int:project_id>/applications/<application_id>/acl/users/<uid>')
class ApiApplicationIdUserIdACL(Resource):
    @admin_api_namespace.marshal_with(success_or_not)
    def delete(self, project_id, application_id, uid):
        user_model = db.session.query(UserModel).filter(UserModel.auth_id == uid).one_or_none()
        if user_model is None:
            raise RekcurdDashboardException("No user found.")

        ApplicationUserRoleModel.query.filter(
            ApplicationUserRoleModel.application_id == application_id,
            ApplicationUserRoleModel.user_id == user_model.user_id).delete()
        db.session.commit()
        db.session.close()
        return {"status": True, "message": "Success."}, 200



import datetime
import tempfile

from flask_restplus import Namespace, fields, Resource, reqparse

from werkzeug.datastructures import FileStorage

from . import DatetimeToTimestamp, status_model
from rekcurd_dashboard.core import RekcurdDashboardClient
from rekcurd_dashboard.data_servers import DataServer
from rekcurd_dashboard.models import db, DataServerModel, DataServerModeEnum, ApplicationModel, ServiceModel, ModelModel
from rekcurd_dashboard.utils import RekcurdDashboardException


model_api_namespace = Namespace('models', description='Model API Endpoint.')
success_or_not = model_api_namespace.model('Status', status_model)
model_model_params = model_api_namespace.model('Model', {
    'model_id': fields.Integer(
        readOnly=True,
        description='Model ID.'
    ),
    'application_id': fields.String(
        readOnly=True,
        description='Application ID.'
    ),
    'filepath': fields.String(
        readOnly=True,
        description='Model file path.',
        example='ml-1234567.model'
    ),
    'register_date': DatetimeToTimestamp(
        readOnly=True,
        description='Register date.'
    ),
    'description': fields.String(
        required=True,
        description='Description.',
        example='This is a sample.'
    )
})


@model_api_namespace.route('/projects/<int:project_id>/applications/<application_id>/models')
class ApiModels(Resource):
    upload_model_parser = reqparse.RequestParser()
    upload_model_parser.add_argument('filepath', location='files',
                                     type=FileStorage, required=True)
    upload_model_parser.add_argument('description', type=str, required=True, location='form')

    @model_api_namespace.marshal_list_with(model_model_params)
    def get(self, project_id: int, application_id: str):
        """get_models"""
        return ModelModel.query.filter_by(application_id=application_id).all()

    @model_api_namespace.marshal_with(success_or_not)
    @model_api_namespace.expect(upload_model_parser)
    def post(self, project_id: int, application_id: str):
        """upload_model"""
        args = self.upload_model_parser.parse_args()
        file: FileStorage = args['filepath']
        description: str = args['description']

        data_server_model: DataServerModel = db.session.query(
            DataServerModel).filter(DataServerModel.project_id == project_id).first_or_404()
        application_model: ApplicationModel = db.session.query(
            ApplicationModel).filter(ApplicationModel.application_id == application_id).first()
        if data_server_model.data_server_mode == DataServerModeEnum.LOCAL:
            """Only if DataServerModeEnum.LOCAL, send file to the server."""
            filepath = "ml-{0:%Y%m%d%H%M%S}.model".format(datetime.datetime.utcnow())
            service_models = db.session.query(
                ServiceModel).filter(ServiceModel.application_id == application_id).all()
            for service_model in service_models:
                rekcurd_dashboard_application = RekcurdDashboardClient(
                    host=service_model.insecure_host, port=service_model.insecure_port,
                    application_name=application_model.application_name,
                    service_level=service_model.service_level, rekcurd_grpc_version=service_model.version)
                response_body = rekcurd_dashboard_application.run_upload_model(filepath, file)
                if not response_body.get("status", True):
                    raise RekcurdDashboardException(response_body.get("message", "Error."))
            response_body = {"status": True, "message": "Success."}
        else:
            """Otherwise, upload file."""
            with tempfile.NamedTemporaryFile() as fp:
                fp.write(file.read())
                data_server = DataServer()
                filepath = data_server.upload_model(data_server_model, application_model, fp.name)
                response_body = {"status": True, "message": "Success."}

        model_model = ModelModel(application_id=application_id, filepath=filepath, description=description)
        db.session.add(model_model)
        db.session.commit()
        db.session.close()
        return response_body


@model_api_namespace.route('/projects/<int:project_id>/applications/<application_id>/models/<int:model_id>')
class ApiModelId(Resource):
    update_config_parser = reqparse.RequestParser()
    update_config_parser.add_argument('description', type=str, required=True, location='form')

    @model_api_namespace.marshal_with(model_model_params)
    def get(self, project_id: int, application_id: str, model_id: int):
        """get_model"""
        return ModelModel.query.filter_by(model_id=model_id).first_or_404()

    @model_api_namespace.marshal_with(success_or_not)
    @model_api_namespace.expect(update_config_parser)
    def patch(self, project_id: int, application_id: str, model_id: int):
        """update_model"""
        args = self.update_config_parser.parse_args()
        description = args['description']

        model_model = db.session.query(ModelModel).filter(ModelModel.model_id==model_id).first_or_404()
        model_model.description = description
        db.session.commit()
        db.session.close()
        return {"status": True, "message": "Success."}

    @model_api_namespace.marshal_with(success_or_not)
    def delete(self, project_id: int, application_id: str, model_id: int):
        """delete_model"""
        model_model = db.session.query(ModelModel).filter(ModelModel.model_id==model_id).one_or_none()
        if model_model is None:
            raise RekcurdDashboardException("No such model_id.")
        num = db.session.query(ServiceModel).filter(ServiceModel.model_id==model_id).count()
        if num > 0:
            raise RekcurdDashboardException("Model is used by some services.")
        # TODO: Delete file.
        db.session.query(ModelModel).filter(ModelModel.model_id==model_id).delete()
        db.session.commit()
        db.session.close()
        return {"status": True, "message": "Success."}



# -*- coding: utf-8 -*-

from .logger_interface import SystemLoggerInterface
from .logger_jsonlogger import JsonSystemLogger
from .logger_fluent import FluentSystemLogger


logger = JsonSystemLogger()



import datetime
import math

from flask_restplus import Namespace, fields, Resource, reqparse, inputs

from . import status_model, apply_rekcurd_to_kubernetes, load_kubernetes_deployment_info
from rekcurd_dashboard.core import RekcurdDashboardClient
from rekcurd_dashboard.models import db, ApplicationModel, ServiceModel
from rekcurd_dashboard.protobuf import rekcurd_pb2


service_deployment_api_namespace = Namespace('service_deployments', description='Service Deployment API Endpoint.')
success_or_not = service_deployment_api_namespace.model('Success', status_model)
service_deployment_params = service_deployment_api_namespace.model('Deployment', {
    'application_name': fields.String(
        readOnly=True,
        description='Application name.'
    ),
    'service_id': fields.String(
        readOnly=True,
        description='Service ID.'
    ),
    'display_name': fields.String(
        required=True,
        description='Display name.',
        example='dev-001'
    ),
    'description': fields.String(
        required=False,
        description='Description.',
        example='This is a sample.'
    ),
    'service_level': fields.String(
        required=True,
        description='Service level. [development/beta/staging/sandbox/production]',
        example='development'
    ),
    'version': fields.String(
        required=False,
        description='Rekcurd gRPC spec version. Default is the latest version.',
        example='v1'
    ),
    'insecure_host': fields.String(
        required=False,
        description='Rekcurd server insecure host. Default is "[::]".',
        example='[::]'
    ),
    'insecure_port': fields.Integer(
        required=False,
        description='Rekcurd server insecure port. Default is "5000".',
        example=5000
    ),
    'replicas_default': fields.Integer(
        required=False,
        description='Number of pod at beginning. Default is "1".',
        example=1
    ),
    'replicas_minimum': fields.Integer(
        required=False,
        description='Minimum number of pod for auto-scaling. Default is "1".',
        example=1
    ),
    'replicas_maximum': fields.Integer(
        required=False,
        description='Maximum number of pod for auto-scaling. Default is "1".',
        example=1
    ),
    'autoscale_cpu_threshold': fields.Integer(
        required=False,
        description='Threshold of CPU usage for auto-scaling. Default "80".',
        example=80
    ),
    'policy_max_surge': fields.Integer(
        required=False,
        description='Maximum number of surged pod when updating. Default "ceil(0.25 * <replicas_default>)".',
        example=1
    ),
    'policy_max_unavailable': fields.Integer(
        required=False,
        description='Maximum number of unavailable pod when updating. Default "floor(0.25 * <replicas_default>)".',
        example=0
    ),
    'policy_wait_seconds': fields.Integer(
        required=False,
        description='Booting second of your service. You MUST specify the accurate number. '
                    'This value become a safety net for your production services. Default is "300".',
        example=300
    ),
    'container_image': fields.String(
        required=True,
        description='Location of your service container image.',
        example='centos:centos7'
    ),
    'service_model_assignment': fields.Integer(
        required=True,
        description='Model ID which is assigned to the service.',
        example=1
    ),
    'service_git_url': fields.String(
        required=False,
        description='URL of your git repository. If you use "rekcurd/rekcurd:tagname" image, '
                    'this field is necessary.',
        example='https://github.com/rekcurd/rekcurd-example.git'
    ),
    'service_git_branch': fields.String(
        required=False,
        description='Name of your git branch. If you use "rekcurd/rekcurd:tagname" image, '
                    'this field is necessary.',
        example='master'
    ),
    'service_boot_script': fields.String(
        required=False,
        description='Booting script for your service. If you use "rekcurd/rekcurd:tagname" image, '
                    'this field is necessary.',
        example='start.sh'
    ),
    'resource_request_cpu': fields.Float(
        required=True,
        description='CPU reservation for your service.',
        example=1.0
    ),
    'resource_request_memory': fields.String(
        required=True,
        description='Memory reservation for your service.',
        example='512Mi'
    ),
    'resource_limit_cpu': fields.Float(
        required=False,
        description='Upper limit of CPU reservation. Default is "resource_request_cpu".',
        example=1.0
    ),
    'resource_limit_memory': fields.String(
        required=False,
        description='Upper limit of memory reservation. Default is "resource_request_memory".',
        example='512Mi'
    ),
    'commit_message': fields.String(
        required=False,
        description='Commit message.',
        example='Initial deployment for "development" env. Default is a commit date.'
    )
})


@service_deployment_api_namespace.route(
    '/projects/<int:project_id>/applications/<application_id>/single_service_registration')
class ApiSingleServiceRegistration(Resource):
    """
    Registration for non-Kubernetes service.
    """
    single_worker_parser = reqparse.RequestParser()
    single_worker_parser.add_argument(
        'display_name', location='form', type=str, required=True, help='Display name. Must be unique.')
    single_worker_parser.add_argument(
        'description', location='form', type=str, required=False, help='Description.')
    single_worker_parser.add_argument(
        'version', location='form', type=str, required=False,
        choices=('v0', 'v1', 'v2'),
        default=rekcurd_pb2.DESCRIPTOR.GetOptions().Extensions[rekcurd_pb2.rekcurd_grpc_proto_version],
        help='Rekcurd gRPC spec version. Default is the latest version.')
    single_worker_parser.add_argument(
        'service_level', location='form', type=str, required=True,
        choices=('development','beta','staging','sandbox','production'),
        help='Service level. [development/beta/staging/sandbox/production].')
    single_worker_parser.add_argument(
        'service_model_assignment', location='form', type=int, required=True,
        help='Model ID which is assigned to the service.')
    single_worker_parser.add_argument(
        'insecure_host', location='form', type=str, default="localhost", required=False,
        help='Rekcurd server host. Default is "localhost".')
    single_worker_parser.add_argument(
        'insecure_port', location='form', type=int, default=5000, required=False,
        help='Rekcurd server port. Default is "5000".')

    @service_deployment_api_namespace.marshal_with(success_or_not)
    @service_deployment_api_namespace.expect(single_worker_parser)
    def post(self, project_id: int, application_id: str):
        """Add non-Kubenetes service."""
        args = self.single_worker_parser.parse_args()
        display_name = args["display_name"]
        description = args["description"]
        service_level = args["service_level"]
        version = args["version"] or rekcurd_pb2.DESCRIPTOR.GetOptions().Extensions[rekcurd_pb2.rekcurd_grpc_proto_version]
        service_model_assignment = args["service_model_assignment"]
        insecure_host = args["insecure_host"]
        insecure_port = args["insecure_port"]

        application_model: ApplicationModel = db.session.query(ApplicationModel).filter(
            ApplicationModel.application_id == application_id).first_or_404()
        rekcurd_dashboard_client = RekcurdDashboardClient(
            host=insecure_host, port=insecure_port, application_name=application_model.application_name,
            service_level=service_level, rekcurd_grpc_version=version)
        service_info = rekcurd_dashboard_client.run_service_info()
        service_id = service_info["service_name"]  # TODO: renaming

        service_model = ServiceModel(
            service_id=service_id, application_id=application_id, display_name=display_name,
            description=description, service_level=service_level, version=version,
            model_id=service_model_assignment, insecure_host=insecure_host,
            insecure_port=insecure_port
        )
        db.session.add(service_model)
        db.session.commit()
        db.session.close()
        return {"status": True, "message": "Success."}


@service_deployment_api_namespace.route('/projects/<int:project_id>/applications/<application_id>/service_deployment')
class ApiServiceDeployment(Resource):
    service_deployment_parser = reqparse.RequestParser()
    service_deployment_parser.add_argument(
        'display_name', location='form', type=str, required=True,
        help='Display name.')
    service_deployment_parser.add_argument(
        'description', location='form', type=str, required=False,
        help='Description.')
    service_deployment_parser.add_argument(
        'version', location='form', type=str, required=False,
        choices=('v0', 'v1', 'v2'),
        default=rekcurd_pb2.DESCRIPTOR.GetOptions().Extensions[rekcurd_pb2.rekcurd_grpc_proto_version],
        help='Rekcurd gRPC spec version. Default is the latest version.')
    service_deployment_parser.add_argument(
        'service_level', location='form', type=str, required=True,
        choices=('development','beta','staging','sandbox','production'),
        help='Service level. [development/beta/staging/sandbox/production].')
    service_deployment_parser.add_argument(
        'insecure_host', location='form', type=str, default="[::]", required=False,
        help='Rekcurd server insecure host. Default is "[::]".')
    service_deployment_parser.add_argument(
        'insecure_port', location='form', type=int, default=5000, required=False,
        help='Rekcurd server insecure port. Default is "5000".')
    service_deployment_parser.add_argument(
        'replicas_default', location='form', type=int, default=1, required=False,
        help='Number of pod at beginning. Default is "1".')
    service_deployment_parser.add_argument(
        'replicas_minimum', location='form', type=int, default=1, required=False,
        help='Minimum number of pod for auto-scaling. Default is "1".')
    service_deployment_parser.add_argument(
        'replicas_maximum', location='form', type=int, default=1, required=False,
        help='Maximum number of pod for auto-scaling. Default is "1".')
    service_deployment_parser.add_argument(
        'autoscale_cpu_threshold', location='form', type=int, default=80, required=False,
        help='Threshold of CPU usage for auto-scaling. Default "80".')
    service_deployment_parser.add_argument(
        'policy_max_surge', location='form', type=int, required=False,
        help='Maximum number of surged pod when updating. Default "ceil(0.25 * <replicas_default>)".')
    service_deployment_parser.add_argument(
        'policy_max_unavailable', location='form', type=int, required=False,
        help='Maximum number of unavailable pod when updating. Default "floor(0.25 * <replicas_default>)".')
    service_deployment_parser.add_argument(
        'policy_wait_seconds', location='form', type=int, default=300, required=False,
        help='Booting second of your service. You MUST specify the accurate number. '
             'This value become a safety net for your production services. Default is "300".')
    service_deployment_parser.add_argument(
        'container_image', location='form', type=str, required=True,
        help='Location of your service container image.')
    service_deployment_parser.add_argument(
        'service_model_assignment', location='form', type=int, required=True,
        help='Model ID which is assigned to the service.')
    service_deployment_parser.add_argument(
        'service_git_url', location='form', type=str, default="", required=False,
        help='URL of your git repository. If you use "rekcurd/rekcurd:tagname" image, this field is necessary.')
    service_deployment_parser.add_argument(
        'service_git_branch', location='form', type=str, default="", required=False,
        help='Name of your git branch. If you use "rekcurd/rekcurd:tagname" image, this field is necessary.')
    service_deployment_parser.add_argument(
        'service_boot_script', location='form', type=str, default="", required=False,
        help='Booting script for your service. If you use "rekcurd/rekcurd:tagname" image, this field is necessary.')
    service_deployment_parser.add_argument(
        'resource_request_cpu', location='form', type=float, default=1.0, required=True,
        help='CPU reservation for your service.')
    service_deployment_parser.add_argument(
        'resource_request_memory', location='form', type=str, default='512Mi', required=True,
        help='Memory reservation for your service.')
    service_deployment_parser.add_argument(
        'resource_limit_cpu', location='form', type=float, required=False,
        help='Upper limit of CPU reservation. Default is "resource_request_cpu".')
    service_deployment_parser.add_argument(
        'resource_limit_memory', location='form', type=str, required=False,
        help='Upper limit of memory reservation. Default is "resource_request_memory".')
    service_deployment_parser.add_argument(
        'debug_mode', location='form', type=inputs.boolean, default=False, required=False,
        help='Debug mode.')

    @service_deployment_api_namespace.marshal_with(success_or_not)
    @service_deployment_api_namespace.expect(service_deployment_parser)
    def post(self, project_id: int, application_id: str):
        """Add Kubenetes service."""
        args = self.service_deployment_parser.parse_args()
        args['commit_message'] = "Update at {0:%Y%m%d%H%M%S}".format(datetime.datetime.utcnow())
        if args['policy_max_surge'] is None:
            args['policy_max_surge'] = math.ceil(0.25 * args['replicas_default'])
        if args['policy_max_unavailable'] is None:
            args['policy_max_unavailable'] = math.floor(0.25 * args['replicas_default'])
        if args['resource_limit_cpu'] is None:
            args['resource_limit_cpu'] = args['resource_request_cpu']
        if args['resource_limit_memory'] is None:
            args['resource_limit_memory'] = args['resource_request_memory']
        apply_rekcurd_to_kubernetes(project_id=project_id, application_id=application_id, **args)
        db.session.commit()
        db.session.close()
        return {"status": True, "message": "Success."}


@service_deployment_api_namespace.route('/projects/<int:project_id>/applications/<application_id>/service_deployment/<service_id>')
class ApiServiceIdDeployment(Resource):
    patch_parser = reqparse.RequestParser()
    patch_parser.add_argument(
        'display_name', location='form', type=str, required=True,
        help='Display name.')
    patch_parser.add_argument(
        'description', location='form', type=str, required=False,
        help='Description.')
    patch_parser.add_argument(
        'version', location='form', type=str, required=True,
        choices=('v0', 'v1', 'v2'),
        default=rekcurd_pb2.DESCRIPTOR.GetOptions().Extensions[rekcurd_pb2.rekcurd_grpc_proto_version],
        help='Rekcurd gRPC spec version. Default is the latest version.')
    patch_parser.add_argument(
        'service_level', location='form', type=str, required=True,
        choices=('development','beta','staging','sandbox','production'),
        help='Service level. [development/beta/staging/sandbox/production].')
    patch_parser.add_argument(
        'insecure_host', location='form', type=str, required=True,
        help='Rekcurd server insecure host. Default is "[::]".')
    patch_parser.add_argument(
        'insecure_port', location='form', type=int, required=True,
        help='Rekcurd server insecure port. Default is "5000".')
    patch_parser.add_argument(
        'replicas_default', location='form', type=int, required=True,
        help='Number of pod at beginning. Default is "1".')
    patch_parser.add_argument(
        'replicas_minimum', location='form', type=int, required=True,
        help='Minimum number of pod for auto-scaling. Default is "1".')
    patch_parser.add_argument(
        'replicas_maximum', location='form', type=int, required=True,
        help='Maximum number of pod for auto-scaling. Default is "1".')
    patch_parser.add_argument(
        'autoscale_cpu_threshold', location='form', type=int, required=True,
        help='Threshold of CPU usage for auto-scaling. Default "80".')
    patch_parser.add_argument(
        'policy_max_surge', location='form', type=int, required=True,
        help='Maximum number of surged pod when updating. Default "ceil(0.25 * <replicas_default>)".')
    patch_parser.add_argument(
        'policy_max_unavailable', location='form', type=int, required=True,
        help='Maximum number of unavailable pod when updating. Default "floor(0.25 * <replicas_default>)".')
    patch_parser.add_argument(
        'policy_wait_seconds', location='form', type=int, required=True,
        help='Booting second of your service. You MUST specify the accurate number. '
             'This value become a safety net for your production services. Default is "300".')
    patch_parser.add_argument(
        'container_image', location='form', type=str, required=True,
        help='Location of your service container image.')
    patch_parser.add_argument(
        'service_model_assignment', location='form', type=int, required=True,
        help='Model ID which is assigned to the service.')
    patch_parser.add_argument(
        'service_git_url', location='form', type=str, required=True,
        help='URL of your git repository. If you use "rekcurd/rekcurd:tagname" image, this field is necessary.')
    patch_parser.add_argument(
        'service_git_branch', location='form', type=str, required=True,
        help='Name of your git branch. If you use "rekcurd/rekcurd:tagname" image, this field is necessary.')
    patch_parser.add_argument(
        'service_boot_script', location='form', type=str, required=True,
        help='Booting script for your service. If you use "rekcurd/rekcurd:tagname" image, this field is necessary.')
    patch_parser.add_argument(
        'resource_request_cpu', location='form', type=float, required=True,
        help='CPU reservation for your service.')
    patch_parser.add_argument(
        'resource_request_memory', location='form', type=str, required=True,
        help='Memory reservation for your service.')
    patch_parser.add_argument(
        'resource_limit_cpu', location='form', type=float, required=True,
        help='Upper limit of CPU reservation. Default is "resource_request_cpu".')
    patch_parser.add_argument(
        'resource_limit_memory', location='form', type=str, required=True,
        help='Upper limit of memory reservation. Default is "resource_request_memory".')
    patch_parser.add_argument(
        'debug_mode', location='form', type=inputs.boolean, default=False, required=False,
        help='Debug mode.')

    @service_deployment_api_namespace.marshal_with(service_deployment_params)
    def get(self, project_id: int, application_id: str, service_id: str):
        """Get Kubernetes deployment info."""
        deployment_info = load_kubernetes_deployment_info(project_id, application_id, service_id)
        service_model: ServiceModel = db.session.query(ServiceModel).filter(
            ServiceModel.service_id == service_id).first_or_404()
        deployment_info["display_name"] = service_model.display_name
        deployment_info["description"] = service_model.description
        return deployment_info

    @service_deployment_api_namespace.marshal_with(success_or_not)
    def put(self, project_id: int, application_id: str, service_id: str):
        """Update Kubernetes deployment info."""
        deployment_info = load_kubernetes_deployment_info(project_id, application_id, service_id)
        service_model: ServiceModel = db.session.query(ServiceModel).filter(
            ServiceModel.service_id == service_id).first_or_404()
        is_updated = False
        if service_model.version != deployment_info["version"]:
            is_updated = True
            service_model.version = deployment_info["version"]
        if service_model.model_id != deployment_info["service_model_assignment"]:
            is_updated = True
            service_model.model_id = deployment_info["service_model_assignment"]
        if service_model.insecure_host != deployment_info["insecure_host"]:
            is_updated = True
            service_model.insecure_host = deployment_info["insecure_host"]
        if service_model.insecure_port != deployment_info["insecure_port"]:
            is_updated = True
            service_model.insecure_port = deployment_info["insecure_port"]
        if is_updated:
            db.session.commit()
        db.session.close()
        return {"status": True, "message": "Success."}

    @service_deployment_api_namespace.marshal_with(success_or_not)
    @service_deployment_api_namespace.expect(patch_parser)
    def patch(self, project_id: int, application_id: str, service_id: str):
        """Rolling update of Kubernetes deployment configurations."""
        args = self.patch_parser.parse_args()
        args['commit_message'] = "Update at {0:%Y%m%d%H%M%S}".format(datetime.datetime.utcnow())
        apply_rekcurd_to_kubernetes(
            project_id=project_id, application_id=application_id, service_id=service_id, **args)

        service_model: ServiceModel = db.session.query(ServiceModel).filter(
            ServiceModel.service_id == service_id).first_or_404()
        is_updated = False
        if service_model.display_name != args["display_name"]:
            is_updated = True
            service_model.version = args["display_name"]
        if service_model.description != args["description"]:
            is_updated = True
            service_model.version = args["description"]
        if service_model.version != args["version"]:
            is_updated = True
            service_model.version = args["version"]
        if service_model.model_id != args["service_model_assignment"]:
            is_updated = True
            service_model.model_id = args["service_model_assignment"]
        if service_model.insecure_host != args["insecure_host"]:
            is_updated = True
            service_model.insecure_host = args["insecure_host"]
        if service_model.insecure_port != args["insecure_port"]:
            is_updated = True
            service_model.insecure_port = args["insecure_port"]
        if is_updated:
            db.session.commit()
        db.session.close()
        return {"status": True, "message": "Success."}



# coding: utf-8


import traceback

from functools import wraps
from flask import jsonify, request, abort
from flask_jwt_simple import JWTManager, create_jwt, get_jwt_identity, jwt_required
from jwt.exceptions import PyJWTError
from flask_jwt_simple.exceptions import InvalidHeaderError, NoAuthorizationError

from .ldap import LdapAuthenticator
from rekcurd_dashboard.utils.exceptions import ProjectUserRoleException, ApplicationUserRoleException
from .authenticator import EmptyAuthenticator
from rekcurd_dashboard.models import (
    db, ApplicationModel, UserModel, ProjectUserRoleModel, ProjectRole, ApplicationUserRoleModel, ApplicationRole
)


class Auth(object):
    def __init__(self):
        self.__enabled = False
        self.authenticator = EmptyAuthenticator()
        self.logger = None

    def is_enabled(self):
        return self.__enabled

    def init_app(self, app, api, auth_conf, logger):
        JWTManager(app)
        self.logger = logger

        if auth_conf is None:
            self.__enabled = False
            self.authenticator = EmptyAuthenticator()
        else:
            self.__enabled = True
            app.config['JWT_SECRET_KEY'] = auth_conf['secret']
            if 'ldap' in auth_conf:
                self.authenticator = LdapAuthenticator(auth_conf['ldap'], logger)

        # Add endpoints
        @app.route('/api/login', methods=['POST'])
        def login():
            params = request.get_json()
            username = params.get('username', None)
            password = params.get('password', None)
            user_info = self.authenticator.auth_user(username, password)
            if user_info is not None:
                user_id = self.user(user_info)
                ret = {'jwt': create_jwt(identity=user_id)}
                return jsonify(ret), 200
            else:
                return jsonify({'message': 'Authentication failed'}), 401

        @app.route('/api/credential', methods=['GET'])
        @jwt_required
        def credential():
            user_id = get_jwt_identity()
            user_model: UserModel = db.session.query(UserModel).filter(UserModel.user_id == user_id).one_or_none()
            if user_model is None:
                abort(404)

            project_roles = user_model.project_roles
            projects = [{'project_id': pr.project_id, 'project_role': pr.project_role.name} for pr in project_roles]
            application_roles = user_model.application_roles
            applications = [{'application_id': ar.application_id, 'application_role': ar.application_role.name}
                            for ar in application_roles]
            # applications which don't have users are also accesssible as admin
            application_ids = db.session.query(ApplicationUserRoleModel.application_id).distinct().all()
            ids = [application_id for application_id, in application_ids]
            public_applications = db.session.query(ApplicationModel).filter(
                ~ApplicationModel.application_id.in_(ids)).all()
            applications += [
                {'application_id': application.application_id, 'role': ApplicationRole.admin.name}
                for application in public_applications]

            return jsonify({'user': user_model.serialize, 'projects': projects, 'applications': applications}), 200

        @api.errorhandler(NoAuthorizationError)
        @api.errorhandler(InvalidHeaderError)
        @api.errorhandler(PyJWTError)
        @api.errorhandler(ApplicationUserRoleException)
        def authorization_error_handler(error):
            self.logger.error(error)
            self.logger.error(traceback.format_exc())
            return {'message': 'Authorization failed'}, 401

    def user(self, user_info):
        uobj = db.session.query(UserModel).filter(UserModel.auth_id == user_info['uid']).one_or_none()
        if uobj is not None:
            return uobj.user_id

        uobj = UserModel(auth_id=user_info['uid'],
                    user_name=user_info['name'])
        db.session.add(uobj)
        db.session.flush()
        db.session.commit()
        user_id = uobj.user_id
        db.session.close()
        return user_id


auth = Auth()


def fetch_project_role(user_id, project_id):
    role: ProjectUserRoleModel = db.session.query(ProjectUserRoleModel).filter(
        ProjectUserRoleModel.project_id == project_id,
        ProjectUserRoleModel.user_id == user_id).one_or_none()
    if role is None:
        return None
    else:
        return role.project_role


def fetch_application_role(user_id, application_id):
    role: ApplicationUserRoleModel = db.session.query(ApplicationUserRoleModel).filter(
        ApplicationUserRoleModel.application_id == application_id,
        ApplicationUserRoleModel.user_id == user_id).one_or_none()
    if role is None:
        # applications which don't have users are also accesssible as owner
        roles = db.session.query(ApplicationUserRoleModel).filter(
            ApplicationUserRoleModel.application_id == application_id).count()
        if roles == 0:
            return ApplicationRole.admin
        else:
            return ApplicationRole.viewer
    else:
        return role.application_role


def auth_required(fn):
    @wraps(fn)
    def wrapper(*args, **kwargs):
        if not auth.is_enabled() or request.path.startswith('/doc/') or request.path.startswith('/api/settings'):
            return fn(*args, **kwargs)
        else:
            @jwt_required
            def run():
                project_id = kwargs.get('project_id')
                if project_id is None:
                    return fn(*args, **kwargs)
                user_id = get_jwt_identity()
                project_user_role = fetch_project_role(user_id, project_id)
                if project_user_role is None:
                    raise ProjectUserRoleException("ProjectUserRoleException")
                application_id = kwargs.get('application_id')
                if application_id is None:
                    if request.method != 'GET' and project_user_role == ProjectRole.member:
                        raise ProjectUserRoleException("ProjectUserRoleException")
                    else:
                        return fn(*args, **kwargs)

                application_user_role = fetch_application_role(user_id, application_id)
                if request.method != 'GET' and application_user_role == ApplicationRole.viewer:
                    raise ApplicationUserRoleException("ApplicationUserRoleException")
                else:
                    return fn(*args, **kwargs)
            return run()
    return wrapper



import datetime
import tempfile
from itertools import chain

from flask import abort
from flask_restplus import Namespace, fields, Resource, reqparse, inputs
from werkzeug.datastructures import FileStorage

from . import status_model
from rekcurd_dashboard.data_servers import DataServer
from rekcurd_dashboard.models import (db, ApplicationModel, ServiceModel, EvaluationModel,
                                      EvaluationResultModel, DataServerModel, DataServerModeEnum)
from rekcurd_dashboard.core import RekcurdDashboardClient
from rekcurd_dashboard.utils import HashUtil, RekcurdDashboardException


evaluation_api_namespace = Namespace('evaluation', description='Evaluation API Endpoint.')
success_or_not = evaluation_api_namespace.model('Success', status_model)
eval_metrics = evaluation_api_namespace.model('Evaluation result', {
    'num': fields.Integer(required=True, description='number of evaluated data'),
    'accuracy': fields.Float(required=True, description='accuracy of evaluation'),
    'fvalue': fields.List(fields.Float, required=True, description='F-value of evaluation'),
    'precision': fields.List(fields.Float, required=True, description='precision of evaluation'),
    'recall': fields.List(fields.Float, required=True, description='recall of evaluation'),
    'option': fields.Raw(),
    'label': fields.Raw(),
    'status': fields.Boolean(required=True),
    'result_id': fields.Integer(required=True, description='ID of evaluation result')
})
eval_data_upload = evaluation_api_namespace.model('Result of uploading evaluation data', {
    'status': fields.Boolean(required=True),
    'evaluation_id': fields.Integer(required=True, description='ID of uploaded data')
})


@evaluation_api_namespace.route('/projects/<int:project_id>/applications/<application_id>/evaluations')
class ApiEvaluations(Resource):
    upload_parser = reqparse.RequestParser()
    upload_parser.add_argument('file', location='files', type=FileStorage, required=True)

    @evaluation_api_namespace.expect(upload_parser)
    @evaluation_api_namespace.marshal_with(eval_data_upload)
    def post(self, project_id: int, application_id: str):
        """update data to be evaluated"""
        args = self.upload_parser.parse_args()
        file = args['file']
        checksum = HashUtil.checksum(file)

        evaluation_model = db.session.query(EvaluationModel).filter(
            EvaluationModel.application_id == application_id,
            EvaluationModel.checksum == checksum).one_or_none()
        if evaluation_model is not None:
            return {"status": True, "evaluation_id": evaluation_model.evaluation_id}

        application_model: ApplicationModel = db.session.query(ApplicationModel).filter(
            ApplicationModel.application_id == application_id).first_or_404()
        data_server_model: DataServerModel = db.session.query(
            DataServerModel).filter(DataServerModel.project_id == project_id).first_or_404()

        if data_server_model.data_server_mode == DataServerModeEnum.LOCAL:
            """Only if DataServerModeEnum.LOCAL, send file to the server."""
            service_model: ServiceModel = db.session.query(ServiceModel).filter(
                ServiceModel.application_id == application_id).first_or_404()
            rekcurd_dashboard_client = RekcurdDashboardClient(
                host=service_model.insecure_host, port=service_model.insecure_port,
                application_name=application_model.application_name,
                service_level=service_model.service_level, rekcurd_grpc_version=service_model.version)
            eval_data_path = "eval-{0:%Y%m%d%H%M%S}.txt".format(datetime.datetime.utcnow())
            response_body = rekcurd_dashboard_client.run_upload_evaluation_data(file, eval_data_path)
            if not response_body['status']:
                raise RekcurdDashboardException('Failed to upload')
        else:
            """Otherwise, upload file."""
            with tempfile.NamedTemporaryFile() as fp:
                fp.write(file.read())
                data_server = DataServer()
                eval_data_path = data_server.upload_evaluation_data(data_server_model, application_model, fp.name)

        evaluation_model = EvaluationModel(
            checksum=checksum, application_id=application_id, data_path=eval_data_path)
        db.session.add(evaluation_model)
        db.session.flush()
        evaluation_id = evaluation_model.evaluation_id
        db.session.commit()
        db.session.close()
        return {"status": True, "evaluation_id": evaluation_id}


@evaluation_api_namespace.route('/projects/<int:project_id>/applications/<application_id>/evaluations/<int:evaluation_id>')
class ApiEvaluationId(Resource):
    @evaluation_api_namespace.marshal_with(success_or_not)
    def delete(self, project_id: int, application_id: str, evaluation_id: int):
        """delete data to be evaluated"""
        eval_query = db.session.query(EvaluationModel).filter(
            EvaluationModel.application_id == application_id,
            EvaluationModel.evaluation_id == evaluation_id)
        evaluation_model = eval_query.one_or_none()
        if evaluation_model is None:
            return {"status": False, "message": "Not Found."}, 404

        data_server_model: DataServerModel = db.session.query(
            DataServerModel).filter(DataServerModel.project_id == project_id).first_or_404()
        data_server = DataServer()
        data_server.delete_file(data_server_model, evaluation_model.data_path)

        eval_query.delete()
        db.session.commit()
        db.session.close()
        return {"status": True, "message": "Success."}


@evaluation_api_namespace.route('/projects/<int:project_id>/applications/<application_id>/evaluate')
class ApiEvaluate(Resource):
    eval_parser = reqparse.RequestParser()
    eval_parser.add_argument('model_id', location='form', type=int, required=True)
    eval_parser.add_argument('evaluation_id', location='form', type=int, required=False)
    eval_parser.add_argument('overwrite', location='form', type=inputs.boolean, required=False)

    @evaluation_api_namespace.expect(eval_parser)
    @evaluation_api_namespace.marshal_with(eval_metrics)
    def post(self, project_id: int, application_id: str):
        """evaluate"""
        args = self.eval_parser.parse_args()
        eval_id = args.get('evaluation_id', None)
        model_id = args.get('model_id')
        if eval_id:
            evaluation_model = EvaluationModel.query.filter_by(
                application_id=application_id,
                evaluation_id=eval_id).first_or_404()
        else:
            # if evaluation_id is not given, use the lastest one.
            evaluation_model = EvaluationModel.query.filter_by(
                application_id=application_id).order_by(EvaluationModel.register_date.desc()).first_or_404()

        # TODO: deploy a temporary service to evaluate, not use an existing service.
        service_model = ServiceModel.query.filter(
            ServiceModel.application_id == application_id,
            ServiceModel.model_id == model_id,
            ServiceModel.service_level != 'production').first()
        if service_model is None:
            raise abort(404, 'The model is not used in any services or used only in production.')

        evaluation_result_model = db.session.query(EvaluationResultModel).filter(
            EvaluationResultModel.model_id == service_model.model_id,
            EvaluationResultModel.evaluation_id == evaluation_model.evaluation_id).one_or_none()
        if evaluation_result_model is not None and args.get('overwrite', False):
            return evaluation_result_model.result

        eval_result_path = "eval-result-{0:%Y%m%d%H%M%S}.pkl".format(datetime.datetime.utcnow())
        application_model: ApplicationModel = db.session.query(ApplicationModel).filter(
            ApplicationModel.application_id == application_id).first_or_404()
        rekcurd_dashboard_client = RekcurdDashboardClient(
            host=service_model.insecure_host, port=service_model.insecure_port, application_name=application_model.application_name,
            service_level=service_model.service_level, rekcurd_grpc_version=service_model.version)
        response_body = rekcurd_dashboard_client.run_evaluate_model(evaluation_model.data_path, eval_result_path)

        if response_body['status']:
            if evaluation_result_model is None:
                evaluation_result_model = EvaluationResultModel(
                    model_id=service_model.model_id,
                    data_path=eval_result_path,
                    evaluation_id=evaluation_model.evaluation_id,
                    result=response_body)
                db.session.add(evaluation_result_model)
            else:
                evaluation_result_model.data_path = eval_result_path
                evaluation_result_model.result = response_body
            db.session.flush()
            response_body = evaluation_result_model.result
            db.session.commit()

        db.session.close()
        return response_body


@evaluation_api_namespace.route('/projects/<int:project_id>/applications/<application_id>/evaluation_results/<int:eval_result_id>')
class ApiEvaluationResults(Resource):

    def get(self, project_id: int, application_id: str, eval_result_id: int):
        """get detailed evaluation result"""
        eval_with_result = db.session.query(EvaluationModel, EvaluationResultModel)\
            .filter(EvaluationModel.application_id == application_id,
                    EvaluationResultModel.evaluation_id == EvaluationModel.evaluation_id,
                    EvaluationResultModel.evaluation_result_id == eval_result_id).one_or_none()
        if eval_with_result is None:
            return {"status": False, "message": "Not Found."}, 404

        application_model: ApplicationModel = db.session.query(ApplicationModel).filter(
            ApplicationModel.application_id == application_id).first_or_404()
        service_model: ServiceModel = db.session.query(ServiceModel).filter(
            ServiceModel.application_id == application_id).first_or_404()
        rekcurd_dashboard_client = RekcurdDashboardClient(
            host=service_model.insecure_host, port=service_model.insecure_port, application_name=application_model.application_name,
            service_level=service_model.service_level, rekcurd_grpc_version=service_model.version)
        evaluation_model = eval_with_result.EvaluationModel
        evaluation_result_model = eval_with_result.EvaluationResultModel

        response_body = list(rekcurd_dashboard_client.run_evaluation_data(
            evaluation_model.data_path, evaluation_result_model.data_path))
        if len(response_body) == 0:
            return {"status": False, "message": "Result Not Found."}, 404

        return {
            'status': all(r['status'] for r in response_body),
            'metrics': evaluation_result_model.result,
            'details': list(chain.from_iterable(r['detail'] for r in response_body))
        }

    @evaluation_api_namespace.marshal_with(success_or_not)
    def delete(self, project_id: int, application_id: str, eval_result_id: int):
        """get detailed evaluation result"""
        eval_with_result = db.session.query(EvaluationModel, EvaluationResultModel)\
            .filter(EvaluationModel.application_id == application_id,
                    EvaluationResultModel.evaluation_id == EvaluationModel.evaluation_id,
                    EvaluationResultModel.evaluation_result_id == eval_result_id).one_or_none()
        if eval_with_result is None:
            return {"status": False, "message": "Not Found."}, 404
        data_server_model: DataServerModel = db.session.query(
            DataServerModel).filter(DataServerModel.project_id == project_id).first_or_404()
        data_server = DataServer()
        evaluation_result_model = eval_with_result.EvaluationResultModel
        data_server.delete_file(data_server_model, evaluation_result_model.data_path)

        db.session.query(EvaluationResultModel).filter(
            EvaluationResultModel.evaluation_result_id == eval_result_id).delete()
        db.session.commit()
        db.session.close()
        return {"status": True, "message": "Success."}



from werkzeug.datastructures import FileStorage


class ProtobufUtil:
    @staticmethod
    def stream_file(f:FileStorage, size:int=4190000) -> bytes:
        for byte in iter(lambda: f.read(size), b''):
            yield byte



import datetime
from .dao import db
from sqlalchemy import (
    Column, Integer, String, DateTime,
    Boolean, Text, UniqueConstraint
)


class ProjectModel(db.Model):
    """
    Project Info
    """
    __tablename__ = 'projects'
    __table_args__ = (
        UniqueConstraint('project_id'),
        UniqueConstraint('display_name'),
        {'mysql_engine': 'InnoDB'}
    )

    project_id = Column(Integer, primary_key=True, autoincrement=True)
    display_name = Column(String(128), nullable=False)
    use_kubernetes = Column(Boolean, nullable=False, default=False)
    description = Column(Text, nullable=True)
    register_date = Column(DateTime, default=datetime.datetime.utcnow, nullable=False)

    @property
    def serialize(self):
        return {
            'project_id': self.project_id,
            'display_name': self.display_name,
            'use_kubernetes': self.use_kubernetes,
            'register_date': self.register_date.strftime('%Y-%m-%d %H:%M:%S'),
            'description': self.description,
        }



import datetime
import enum
from .dao import db
from sqlalchemy import (
    Column, Integer, String, DateTime, Enum,
    Boolean, UniqueConstraint, ForeignKey
)
from sqlalchemy.orm import relationship
from sqlalchemy.orm import backref


class DataServerModeEnum(enum.Enum):
    """
    Rekcurd data server options.
    """
    LOCAL = 'local'
    CEPH_S3 = 'ceph_s3'
    AWS_S3 = 'aws_s3'
    GCS = 'gcs'
    PV = 'pv'  # Need to set env as 'local' for Rekcurd worker

    @classmethod
    def to_enum(cls, mode: str):
        if cls.LOCAL.value == mode:
            return cls.LOCAL
        elif cls.CEPH_S3.value == mode:
            return cls.CEPH_S3
        elif cls.AWS_S3.value == mode:
            return cls.AWS_S3
        elif cls.GCS.value == mode:
            return cls.GCS
        elif cls.PV.value == mode:
            return cls.PV
        else:
            raise ValueError("'{}' is not supported as ModelModeEnum".format(mode))


class DataServerModel(db.Model):
    """
    Data Server Info
    """
    __tablename__ = 'data_servers'
    __table_args__ = (
        UniqueConstraint('project_id'),
        {'mysql_engine': 'InnoDB'}
    )

    project_id = Column(Integer, ForeignKey('projects.project_id', ondelete="CASCADE"), primary_key=True)
    data_server_mode = Column(Enum(DataServerModeEnum), nullable=False, default=DataServerModeEnum.LOCAL)
    ceph_access_key = Column(String(128), nullable=False, default='')
    ceph_secret_key = Column(String(128), nullable=False, default='')
    ceph_host = Column(String(512), nullable=False, default='')
    ceph_port = Column(Integer, nullable=False, default=80)
    ceph_is_secure = Column(Boolean, nullable=False, default=False)
    ceph_bucket_name = Column(String(128), nullable=False, default='')
    aws_access_key = Column(String(128), nullable=False, default='')
    aws_secret_key = Column(String(128), nullable=False, default='')
    aws_bucket_name = Column(String(128), nullable=False, default='')
    # TODO: GCP
    # TODO: PV
    register_date = Column(DateTime, default=datetime.datetime.utcnow, nullable=False)

    project = relationship(
        'ProjectModel', innerjoin=True,
        backref=backref("data_servers", cascade="all, delete-orphan", passive_deletes=True))



from flask_jwt_simple import get_jwt_identity
from flask_restplus import Namespace, fields, Resource, reqparse, inputs

from . import api, status_model, update_kubernetes_deployment_info
from rekcurd_dashboard.utils import RekcurdDashboardException
from rekcurd_dashboard.models import db, ProjectModel, ProjectUserRoleModel, ProjectRole, KubernetesModel
from rekcurd_dashboard.apis import DatetimeToTimestamp


project_api_namespace = Namespace('projects', description='Project API Endpoint.')
success_or_not = project_api_namespace.model('Status', status_model)
project_model_params = project_api_namespace.model('Project', {
    'project_id': fields.Integer(
        readOnly=True,
        description='Project ID.'
    ),
    'display_name': fields.String(
        required=True,
        description='Project name.',
        example='Sample project.'
    ),
    'use_kubernetes': fields.Boolean(
        required=True,
        description='Do you use Kubernetes?',
        example='False'
    ),
    'description': fields.String(
        required=False,
        description='Description.',
        example='This is a sample.'
    ),
    'register_date': DatetimeToTimestamp(
        readOnly=True,
        description='Register date.'
    )
})


@project_api_namespace.route('/projects')
class ApiProjects(Resource):
    project_model_parser = reqparse.RequestParser()
    project_model_parser.add_argument('display_name', type=str, required=True, location='form', default='sample', help='Sample project.')
    project_model_parser.add_argument('use_kubernetes', type=inputs.boolean, required=True, location='form', default=False, help='Do you use Kubernetes?')
    project_model_parser.add_argument('description', type=str, required=False, location='form', help='Description.')

    @project_api_namespace.marshal_list_with(project_model_params)
    def get(self):
        """get_projects"""
        return ProjectModel.query.all()

    @project_api_namespace.marshal_with(success_or_not)
    @project_api_namespace.expect(project_model_parser)
    def post(self):
        """add_non-kube_worker"""
        args = self.project_model_parser.parse_args()
        display_name = args['display_name']
        description = args['description']
        use_kubernetes = args['use_kubernetes']

        project_model = db.session.query(ProjectModel).filter(
            ProjectModel.display_name == display_name).one_or_none()
        if project_model is None:
            project_model = ProjectModel(
                display_name=display_name, description=description, use_kubernetes=use_kubernetes)
            db.session.add(project_model)
            db.session.flush()
        else:
            raise RekcurdDashboardException("Project name is duplicated. \"{}\"".format(display_name))

        if api.dashboard_config.IS_ACTIVATE_AUTH:
            user_id = get_jwt_identity()
            project_user_role_model = ProjectUserRoleModel(
                project_id=project_model.project_id,
                user_id=user_id,
                project_role=ProjectRole.admin.name)
            db.session.add(project_user_role_model)
            db.session.flush()
        db.session.commit()
        db.session.close()
        return {"status": True, "message": "Success."}


@project_api_namespace.route('/projects/<int:project_id>')
class ApiProjectId(Resource):
    project_model_parser = reqparse.RequestParser()
    project_model_parser.add_argument('display_name', type=str, required=False, location='form', help='Sample project.')
    project_model_parser.add_argument('use_kubernetes', type=bool, required=False, location='form', help='Do you use Kubernetes?')
    project_model_parser.add_argument('description', type=str, required=False, location='form', help='Description.')

    @project_api_namespace.marshal_with(project_model_params)
    def get(self, project_id:int):
        """get_project"""
        return ProjectModel.query.filter_by(project_id=project_id).first_or_404()

    @project_api_namespace.marshal_with(success_or_not)
    @project_api_namespace.expect(project_model_parser)
    def patch(self, project_id:int):
        """update_project"""
        args = self.project_model_parser.parse_args()
        display_name = args['display_name']
        description = args['description']
        use_kubernetes = args['use_kubernetes']

        project_model = db.session.query(ProjectModel).filter(ProjectModel.project_id == project_id).first_or_404()
        is_update = False
        if display_name is not None:
            project_model.display_name = display_name
            is_update = True
        if description is not None:
            project_model.description = description
            is_update = True
        if use_kubernetes is not None:
            project_model.use_kubernetes = use_kubernetes
            is_update = True
        if is_update:
            db.session.commit()
        db.session.close()
        return {"status": True, "message": "Success."}

    @project_api_namespace.marshal_with(success_or_not)
    def put(self, project_id: int):
        """update to the latest Kubernetes deployment info."""
        kubernetes_model = KubernetesModel.query.filter_by(project_id=project_id).first()
        update_kubernetes_deployment_info(kubernetes_model)
        db.session.commit()
        db.session.close()
        return {"status": True, "message": "Success."}



import uuid

from flask_restplus import Namespace, fields, Resource, reqparse
from flask_jwt_simple import get_jwt_identity

from werkzeug.datastructures import FileStorage

from . import (
    api, DatetimeToTimestamp, status_model, update_kubernetes_deployment_info,
    save_kubernetes_access_file, remove_kubernetes_access_file, backup_kubernetes_deployment,
    backup_istio_routing
)
from rekcurd_dashboard.models import (
    db, KubernetesModel, ProjectUserRoleModel, ProjectRole, ApplicationModel, ServiceModel
)
from rekcurd_dashboard.utils import ProjectUserRoleException


kubernetes_api_namespace = Namespace('kubernetes', description='Kubernetes API Endpoint.')
success_or_not = kubernetes_api_namespace.model('Success', status_model)
true_or_false = kubernetes_api_namespace.model('True or False', {
    'is_target': fields.Boolean(
        readOnly=True,
        description='True or False'
    )
})
kubernetes_model_params = kubernetes_api_namespace.model('Kubernetes', {
    'kubernetes_id': fields.Integer(
        readOnly=True,
        description='Kubernetes cluster ID.'
    ),
    'project_id': fields.Integer(
        readOnly=True,
        description='Project ID.'
    ),
    'display_name': fields.String(
        required=True,
        description='DisplayName.',
        example='cluster-1'
    ),
    'description': fields.String(
        required=False,
        description='Description.',
        example='This is a sample.'
    ),
    'config_path': fields.String(
        readOnly=True,
        description='Kubernetes configuration file path.',
        example='/kube-config/kube-1.config'
    ),
    'exposed_host': fields.String(
        required=True,
        description='Ingress host or External IP.',
        example='127.0.0.1'
    ),
    'exposed_port': fields.Integer(
        required=True,
        description='Ingress port or NodePort.',
        example='33000'
    ),
    'register_date': DatetimeToTimestamp(
        readOnly=True,
        description='Register date.'
    )
})


@kubernetes_api_namespace.route('/projects/<int:project_id>/kubernetes')
class ApiKubernetes(Resource):
    kubernetes_model_parser = reqparse.RequestParser()
    kubernetes_model_parser.add_argument(
        'file', location='files', type=FileStorage, required=True,
        help='Kubernetes access_token configuration file.')
    kubernetes_model_parser.add_argument(
        'display_name', location='form', type=str, required=True,
        help="Must be unique.")
    kubernetes_model_parser.add_argument(
        'description', location='form', type=str, required=False,
        help='Description.')
    kubernetes_model_parser.add_argument(
        'exposed_host', location='form', type=str, required=True,
        help='Ingress host or External IP. e.g. "127.0.0.1", "example.com"')
    kubernetes_model_parser.add_argument(
        'exposed_port', location='form', type=int, required=True,
        help='Ingress port or NodePort. e.g. "33000"')

    @kubernetes_api_namespace.marshal_list_with(kubernetes_model_params)
    def get(self, project_id: int):
        """get_kubernetes"""
        if api.dashboard_config.IS_ACTIVATE_AUTH:
            user_id = get_jwt_identity()
            role: ProjectUserRoleModel = db.session.query(ProjectUserRoleModel).filter(
                ProjectUserRoleModel.project_id == project_id,
                ProjectUserRoleModel.user_id == user_id).one_or_none()
            if role is None:
                raise ProjectUserRoleException("Invalid access.")
        return KubernetesModel.query.filter_by(project_id=project_id).all()

    @kubernetes_api_namespace.marshal_with(success_or_not)
    @kubernetes_api_namespace.expect(kubernetes_model_parser)
    def post(self, project_id: int):
        """add_kubernetes"""
        args = self.kubernetes_model_parser.parse_args()
        config_path = "{0}.config".format(uuid.uuid4().hex)
        file = args['file']
        display_name = args['display_name']
        description = args['description']
        exposed_host = args['exposed_host']
        exposed_port = args['exposed_port']
        kubernetes_model = KubernetesModel(
            project_id=project_id, display_name=display_name, description=description,
            config_path=config_path, exposed_host=exposed_host, exposed_port=exposed_port)
        db.session.add(kubernetes_model)
        db.session.flush()
        save_kubernetes_access_file(file, config_path)
        try:
            update_kubernetes_deployment_info(kubernetes_model)
            db.session.commit()
            db.session.close()
        except Exception as error:
            remove_kubernetes_access_file(config_path)
            raise error
        return {"status": True, "message": "Success."}


@kubernetes_api_namespace.route('/projects/<int:project_id>/backup')
class ApiKubernetesBackup(Resource):
    @kubernetes_api_namespace.marshal_with(success_or_not)
    def post(self, project_id: int):
        """backup Kubernetes deployment"""
        kubernetes_models = KubernetesModel.query.filter_by(project_id=project_id).all()
        if kubernetes_models:
            kubernetes_model = kubernetes_models[0]
            for application_model in ApplicationModel.query.filter_by(project_id=project_id).all():
                service_level = None
                for service_model in ServiceModel.query.filter_by(application_id=application_model.application_id).all():
                    service_level = service_model.service_level
                    backup_kubernetes_deployment(kubernetes_model, application_model, service_model)
                backup_istio_routing(kubernetes_model, application_model, service_level)
        return {"status": True, "message": "Success."}


@kubernetes_api_namespace.route('/projects/<int:project_id>/kubernetes/<int:kubernetes_id>')
class ApiKubernetesId(Resource):
    kubernetes_model_parser = reqparse.RequestParser()
    kubernetes_model_parser.add_argument(
        'file', location='files', type=FileStorage, required=False,
        help='Kubernetes access_token configuration file.')
    kubernetes_model_parser.add_argument(
        'display_name', location='form', type=str, required=False,
        help="Must be unique. If empty, automatically generated.")
    kubernetes_model_parser.add_argument(
        'description', location='form', type=str, required=False,
        help='Description.')
    kubernetes_model_parser.add_argument(
        'exposed_host', location='form', type=str, required=False,
        help='Ingress host or External IP. e.g. "127.0.0.1", "example.com"')
    kubernetes_model_parser.add_argument(
        'exposed_port', location='form', type=int, required=False,
        help='Ingress port or NodePort. e.g. "33000"')

    @kubernetes_api_namespace.marshal_with(kubernetes_model_params)
    def get(self, project_id: int, kubernetes_id: int):
        """get_kubernetes_id"""
        if api.dashboard_config.IS_ACTIVATE_AUTH:
            user_id = get_jwt_identity()
            role: ProjectUserRoleModel = db.session.query(ProjectUserRoleModel).filter(
                ProjectUserRoleModel.project_id == project_id,
                ProjectUserRoleModel.user_id == user_id).one_or_none()
            if role is None or role.project_role != ProjectRole.admin:
                raise ProjectUserRoleException("Invalid access.")
        return KubernetesModel.query.filter_by(kubernetes_id=kubernetes_id).first_or_404()

    @kubernetes_api_namespace.marshal_with(success_or_not)
    @kubernetes_api_namespace.expect(kubernetes_model_parser)
    def patch(self, project_id: int, kubernetes_id: int):
        """update_kubernetes_id"""
        args = self.kubernetes_model_parser.parse_args()
        config_path = "{0}.config".format(uuid.uuid4().hex)
        file = args['file']
        display_name = args['display_name']
        description = args['description']
        exposed_host = args['exposed_host']
        exposed_port = args['exposed_port']

        kubernetes_model: KubernetesModel = db.session.query(KubernetesModel).filter(
            KubernetesModel.kubernetes_id==kubernetes_id).first_or_404()
        prev_config_path = kubernetes_model.config_path
        is_update = False
        if display_name is not None:
            is_update = True
            kubernetes_model.display_name = display_name
        if description is not None:
            is_update = True
            kubernetes_model.description = description
        if exposed_host is not None:
            is_update = True
            kubernetes_model.exposed_host = exposed_host
        if exposed_port is not None:
            is_update = True
            kubernetes_model.exposed_port = exposed_port
        if file is not None:
            is_update = True
            kubernetes_model.config_path = config_path
            save_kubernetes_access_file(file, config_path)
            try:
                update_kubernetes_deployment_info(kubernetes_model)
                remove_kubernetes_access_file(prev_config_path)
            except Exception as error:
                remove_kubernetes_access_file(config_path)
                raise error
        if is_update:
            db.session.commit()
            db.session.close()
        return {"status": True, "message": "Success."}

    @kubernetes_api_namespace.marshal_with(success_or_not)
    def delete(self, project_id: int, kubernetes_id: int):
        """delete_kubernetes_id"""
        db.session.query(KubernetesModel).filter(KubernetesModel.kubernetes_id == kubernetes_id).delete()
        db.session.commit()
        db.session.close()
        return {"status": True, "message": "Success."}



# coding: utf-8


from .exceptions import RekcurdDashboardException, ProjectUserRoleException, ApplicationUserRoleException
from .hash_util import HashUtil
from .protobuf_util import ProtobufUtil
from .rekcurd_dashboard_config import RekcurdDashboardConfig



# -*- coding: utf-8 -*-


from rekcurd_dashboard.core import create_app


def db_handler(args: dict):
    import sys
    from flask_script import Manager
    from flask_migrate import Migrate, MigrateCommand
    from rekcurd_dashboard.models import db

    tmp = sys.argv[1:]
    sys.argv = [sys.argv[0]]
    for i, t in enumerate(tmp):
        if t == 'db':
            sys.argv.append(t)
            sys.argv.append(tmp[i+1])
            break
    app, _ = create_app(config_file=args["settings"], logger_file=args["logger"], **args)
    migrate = Migrate(app, db)
    manager = Manager(app)
    manager.add_command('db', MigrateCommand)
    manager.run()



# coding: utf-8


import yaml
import os
import json
from rekcurd_dashboard.protobuf import rekcurd_pb2


class RekcurdDashboardConfig:
    """
    Rekcurd dashboard configurations.
    """
    __SERVICE_DEFAULT_PORT: int = 18080
    REKCURD_GRPC_VERSION: str = rekcurd_pb2.DESCRIPTOR.GetOptions().Extensions[rekcurd_pb2.rekcurd_grpc_proto_version]

    __TEST_MODE: bool = None
    KUBERNETES_MODE: str = None
    DEBUG_MODE: bool = None
    SERVICE_PORT: int = __SERVICE_DEFAULT_PORT
    DIR_KUBE_CONFIG: str = None
    SQLALCHEMY_DATABASE_URI: str = None
    IS_ACTIVATE_AUTH: bool = None
    AUTH_CONFIG: dict = None

    def __init__(self, config_file: str = None):
        self.__TEST_MODE = os.getenv("DASHBOARD_TEST_MODE", "False").lower() == 'true'
        self.KUBERNETES_MODE = os.getenv("DASHBOARD_KUBERNETES_MODE")
        if self.KUBERNETES_MODE is None:
            self.__load_from_file(config_file)
        else:
            self.__load_from_env()

    def set_configurations(
            self, debug_mode: bool = None, port: int = None,
            kube_config_dir: str = None,
            db_mode: str = None, db_host: str = None, db_port: int = None,
            db_name: str = None, db_username: str = None, db_password: str = None,
            **options):
        self.DEBUG_MODE = debug_mode if debug_mode is not None else self.DEBUG_MODE
        self.SERVICE_PORT = int(port or self.SERVICE_PORT)
        self.DIR_KUBE_CONFIG = kube_config_dir or self.DIR_KUBE_CONFIG
        self.SQLALCHEMY_DATABASE_URI = \
            self.__create_db_uri(db_mode, db_host, db_port, db_name, db_username, db_password) or \
            self.SQLALCHEMY_DATABASE_URI
        # TODO: Auth

    def __load_from_file(self, config_file: str):
        if config_file is not None:
            with open(config_file, 'r') as f:
                config = yaml.load(f)
        else:
            config = dict()
        self.DEBUG_MODE = config.get("debug", False)
        self.SERVICE_PORT = config.get("port", self.SERVICE_PORT)
        self.DIR_KUBE_CONFIG = config.get("kube_config_dir", "kube-config")
        config_db = config.get("db", dict())
        db_mode = config_db.get("mode", "sqlite")
        config_db_mysql = config_db.get("mysql", dict())
        db_host = config_db_mysql.get("host")
        db_port = config_db_mysql.get("port")
        db_name = config_db_mysql.get("dbname")
        db_username = config_db_mysql.get("username")
        db_password = config_db_mysql.get("password")
        self.SQLALCHEMY_DATABASE_URI = \
            self.__create_db_uri(db_mode, db_host, db_port, db_name, db_username, db_password)
        if 'auth' in config:
            self.IS_ACTIVATE_AUTH = True
            self.AUTH_CONFIG = config['auth']
        else:
            self.IS_ACTIVATE_AUTH = False
            self.AUTH_CONFIG = None

    def __load_from_env(self):
        self.DEBUG_MODE = os.getenv("DASHBOARD_DEBUG_MODE", "False").lower() == 'true'
        self.SERVICE_PORT = int(os.getenv("DASHBOARD_SERVICE_PORT", "{}".format(self.__SERVICE_DEFAULT_PORT)))
        self.DIR_KUBE_CONFIG = os.getenv("DASHBOARD_KUBE_DATADIR")
        db_mode = os.getenv('DASHBOARD_DB_MODE')
        db_host = os.getenv('DASHBOARD_DB_MYSQL_HOST')
        db_port = os.getenv('DASHBOARD_DB_MYSQL_PORT')
        db_name = os.getenv('DASHBOARD_DB_MYSQL_DBNAME')
        db_username = os.getenv('DASHBOARD_DB_MYSQL_USERNAME')
        db_password = os.getenv('DASHBOARD_DB_MYSQL_PASSWORD')
        self.SQLALCHEMY_DATABASE_URI = \
            self.__create_db_uri(db_mode, db_host, db_port, db_name, db_username, db_password)
        if os.getenv('DASHBOARD_IS_AUTH', 'False').lower() == 'true':
            self.IS_ACTIVATE_AUTH = True
            self.AUTH_CONFIG = {
                'secret': os.getenv('DASHBOARD_LDAP_SECRET'),
                'ldap': {
                    'host': os.getenv('DASHBOARD_LDAP_HOST'),
                    'port': os.getenv('DASHBOARD_LDAP_PORT'),
                    'bind_dn': os.getenv('DASHBOARD_LDAP_BIND_DN'),
                    'bind_password': os.getenv('DASHBOARD_LDAP_BIND_PASSWORD'),
                    'search_filter': os.getenv('DASHBOARD_LDAP_SEARCH_FILTER'),
                    'search_base_dns': json.loads(os.getenv('DASHBOARD_LDAP_SEARCH_BASE_DNS')),
                }
            }
        else:
            self.IS_ACTIVATE_AUTH = False
            self.AUTH_CONFIG = None

    def __create_db_uri(self, db_mode, db_host, db_port, db_name, db_username, db_password):
        if self.__TEST_MODE:
            from sqlalchemy.engine import Engine
            from sqlalchemy import event

            @event.listens_for(Engine, "connect")
            def set_sqlite_pragma(dbapi_connection, connection_record):
                cursor = dbapi_connection.cursor()
                cursor.execute("PRAGMA foreign_keys=ON")
                cursor.close()
            return f'sqlite:///db.test.sqlite3'
        elif db_mode is None:
            return None
        elif db_mode == "sqlite":
            from sqlalchemy.engine import Engine
            from sqlalchemy import event

            @event.listens_for(Engine, "connect")
            def set_sqlite_pragma(dbapi_connection, connection_record):
                cursor = dbapi_connection.cursor()
                cursor.execute("PRAGMA foreign_keys=ON")
                cursor.close()
            return f'sqlite:///db.sqlite3'
        elif db_mode == "mysql" and db_host and db_port and db_name and db_username and db_password:
            return f'mysql+pymysql://{db_username}:{db_password}@{db_host}:{db_port}/{db_name}?charset=utf8'
        else:
            raise TypeError("Invalid DB configurations.")



import enum
from .dao import db
from sqlalchemy import (
    Column, Integer,
    Enum, String,
    ForeignKey,
    UniqueConstraint
)
from sqlalchemy.orm import relationship
from sqlalchemy.orm import backref


class ApplicationRole(enum.Enum):
    admin = 1
    editor = 2
    viewer = 3


class ApplicationUserRoleModel(db.Model):
    """
    Application-User Role
    """
    __tablename__ = 'application_user_roles'
    __table_args__ = (
        UniqueConstraint('application_id', 'user_id'),
        {'mysql_engine': 'InnoDB'}
    )

    id = Column(Integer, primary_key=True, autoincrement=True)
    user_id = Column(Integer, ForeignKey('users.user_id', ondelete="CASCADE"), nullable=False)
    application_id = Column(String, ForeignKey('applications.application_id', ondelete="CASCADE"), nullable=True)
    application_role = Column(Enum(ApplicationRole), nullable=False, default=ApplicationRole.viewer)

    application = relationship(
        'ApplicationModel', lazy='joined', innerjoin=True,
        backref=backref("application_user_roles", cascade="all, delete-orphan", passive_deletes=True))
    user = relationship(
        'UserModel',
        backref=backref("application_user_roles", cascade="all, delete-orphan", passive_deletes=True))



import datetime
from .dao import db
from sqlalchemy import (
    Column, Integer, String, DateTime,
    Text, UniqueConstraint, ForeignKey
)
from sqlalchemy.orm import relationship
from sqlalchemy.orm import backref


class KubernetesModel(db.Model):
    """
    Kubernetes Access Info
    """
    __tablename__ = 'kubernetes'
    __table_args__ = (
        UniqueConstraint('kubernetes_id'),
        UniqueConstraint('config_path'),
        UniqueConstraint('display_name'),
        UniqueConstraint('exposed_host', 'exposed_port'),
        {'mysql_engine': 'InnoDB'}
    )

    kubernetes_id = Column(Integer, primary_key=True, autoincrement=True)
    project_id = Column(Integer, ForeignKey('projects.project_id', ondelete="CASCADE"), nullable=False)
    display_name = Column(String(128), nullable=False)
    description = Column(Text, nullable=True)
    config_path = Column(String(512), nullable=False)
    exposed_host = Column(String(512), nullable=False)
    exposed_port = Column(Integer, nullable=False)
    register_date = Column(DateTime, default=datetime.datetime.utcnow, nullable=False)

    projects = relationship(
        'ProjectModel', innerjoin=True,
        backref=backref("kubernetes", cascade="all, delete-orphan", passive_deletes=True))

    @property
    def serialize(self):
        return {
            'kubernetes_id': self.kubernetes_id,
            'project_id': self.project_id,
            'display_name': self.display_name,
            'description': self.description,
            'config_path': self.config_path,
            'exposed_host': self.exposed_host,
            'exposed_port': self.exposed_port,
            'register_date': self.register_date.strftime('%Y-%m-%d %H:%M:%S'),
        }



# -*- coding: utf-8 -*-


import argparse

from rekcurd_dashboard import _version

from .server_handler import server_handler
from .db_handler import db_handler


def create_parser():
    parser = argparse.ArgumentParser(description='rekcurd_dashboard command')
    parser.add_argument(
        '--version', '-v', action='version', version=_version.__version__)
    subparsers = parser.add_subparsers()

    # server
    parser_server = subparsers.add_parser(
        'server', help='see `rekcurd_dashboard server -h`')
    parser_server.add_argument(
        '-H', '--host', required=False, help='host', default='0.0.0.0')
    parser_server.add_argument(
        '-p', '--port', required=False, type=int, help='port')
    parser_server.add_argument(
        '--settings', required=False, help='settings YAML. See https://github.com/rekcurd/dashboard/blob/master/rekcurd_dashboard/template/settings.yml-tpl')
    parser_server.add_argument(
        '--logger', required=False, help='Python file of your custom logger. Need to inherit "logger_interface.py". See https://github.com/rekcurd/dashboard/blob/master/rekcurd_dashboard/logger/logger_interface.py')
    parser_server.add_argument(
        '--debug', required=False, type=bool, help='Debug mode.')
    parser_server.add_argument(
        '--kube_config_dir', required=False, help='Directory of kube_config files.')
    parser_server.add_argument(
        '--db_mode', required=False, help='Dashboard DB mode. One of [sqlite/mysql]. Default "sqlite".')
    parser_server.add_argument(
        '--db_host', required=False, help='Dashboard MySQL host. Necessary if "--db_mode mysql".')
    parser_server.add_argument(
        '--db_port', required=False, type=int, help='Dashboard MySQL port. Necessary if "--db_mode mysql".')
    parser_server.add_argument(
        '--db_name', required=False, help='Dashboard MySQL DB name. Necessary if "--db_mode mysql".')
    parser_server.add_argument(
        '--db_username', required=False, help='Dashboard MySQL username. Necessary if "--db_mode mysql".')
    parser_server.add_argument(
        '--db_password', required=False, help='Dashboard MySQL password. Necessary if "--db_mode mysql".')
    parser_server.set_defaults(handler=server_handler)

    # db
    parser_db = subparsers.add_parser(
        'db', help='see `rekcurd_dashboard db -h`')
    parser_db.add_argument(
        'type', choices=['init', 'revision', 'migrate', 'edit', 'merge',
                         'upgrade', 'downgrade', 'show', 'history', 'heads',
                         'branches', 'current', 'stamp'])
    parser_db.add_argument(
        '--settings', required=False, help='settings YAML. See https://github.com/rekcurd/dashboard/blob/master/rekcurd_dashboard/template/settings.yml-tpl')
    parser_db.add_argument(
        '--logger', required=False, help='Python file of your custom logger. Need to inherit "logger_interface.py". See https://github.com/rekcurd/dashboard/blob/master/rekcurd_dashboard/logger/logger_interface.py')
    parser_db.add_argument(
        '--db_mode', required=False, help='Dashboard DB mode. One of [sqlite/mysql]. Default "sqlite".')
    parser_db.add_argument(
        '--db_host', required=False, help='Dashboard MySQL host. Necessary if "--db_mode mysql".')
    parser_db.add_argument(
        '--db_port', required=False, type=int, help='Dashboard MySQL port. Necessary if "--db_mode mysql".')
    parser_db.add_argument(
        '--db_name', required=False, help='Dashboard MySQL DB name. Necessary if "--db_mode mysql".')
    parser_db.add_argument(
        '--db_username', required=False, help='Dashboard MySQL username. Necessary if "--db_mode mysql".')
    parser_db.add_argument(
        '--db_password', required=False, help='Dashboard MySQL password. Necessary if "--db_mode mysql".')
    parser_db.set_defaults(handler=db_handler)

    return parser


def main() -> None:
    parser = create_parser()
    args = parser.parse_args()

    if hasattr(args, 'handler'):
        args.handler(vars(args))
    else:
        parser.print_help()


if __name__ == '__main__':
    main()



import enum
from .dao import db
from sqlalchemy import (
    Column, Integer,
    Enum,
    ForeignKey,
    UniqueConstraint
)
from sqlalchemy.orm import relationship
from sqlalchemy.orm import backref


class ProjectRole(enum.Enum):
    admin = 1
    member = 2


class ProjectUserRoleModel(db.Model):
    """
    Project-User Role
    """
    __tablename__ = 'project_user_roles'
    __table_args__ = (
        UniqueConstraint('project_id', 'user_id'),
        {'mysql_engine': 'InnoDB'}
    )

    id = Column(Integer, primary_key=True, autoincrement=True)
    user_id = Column(Integer, ForeignKey('users.user_id', ondelete="CASCADE"), nullable=False)
    project_id = Column(Integer, ForeignKey('projects.project_id', ondelete="CASCADE"), nullable=False)
    project_role = Column(Enum(ProjectRole), nullable=False)

    project = relationship(
        'ProjectModel', lazy='joined', innerjoin=True,
        backref=backref("project_user_roles", cascade="all, delete-orphan", passive_deletes=True))
    user = relationship(
        'UserModel',
        backref=backref("project_user_roles", cascade="all, delete-orphan", passive_deletes=True))



import datetime
import json
from .dao import db
from sqlalchemy import (
    Column, Integer, DateTime, Text,
    UniqueConstraint, ForeignKey, String
)
from sqlalchemy.orm import relationship
from sqlalchemy.orm import backref


class EvaluationResultModel(db.Model):
    """
    Evaluation Info
    """
    __tablename__ = 'evaluation_results'
    __table_args__ = (
        UniqueConstraint('evaluation_result_id'),
        {'mysql_engine': 'InnoDB'}
    )

    evaluation_result_id = Column(Integer, primary_key=True, autoincrement=True)
    model_id = Column(Integer, ForeignKey('models.model_id', ondelete="CASCADE"), nullable=False)
    data_path = Column(String(512), nullable=False)
    evaluation_id = Column(Integer, ForeignKey('evaluations.evaluation_id', ondelete="CASCADE"), nullable=False)
    _result = Column(Text, nullable=False)
    register_date = Column(DateTime, default=datetime.datetime.utcnow, nullable=False)

    model = relationship(
        'ModelModel', innerjoin=True,
        backref=backref("evaluation_results", cascade="all, delete-orphan", passive_deletes=True))
    evaluations = relationship(
        'EvaluationModel', lazy='select', innerjoin=True,
        backref=backref("evaluation_results", cascade="all, delete-orphan", passive_deletes=True))

    @property
    def result(self):
        res = json.loads(self._result)
        res['result_id'] = self.evaluation_result_id
        return res

    @result.setter
    def result(self, value):
        if isinstance(value, dict):
            res = json.dumps(value)
        else:
            res = value
        self._result = res

    @property
    def serialize(self):
        return {
            'evaluation_result_id': self.evaluation_result_id,
            'model_id': self.model_id,
            'data_path': self.data_path,
            'evaluation_id': self.evaluation_id,
            'result': self.result,
            'register_date': self.register_date.strftime('%Y-%m-%d %H:%M:%S'),
        }



from abc import ABCMeta, abstractmethod


class Authenticator(metaclass=ABCMeta):
    @abstractmethod
    def auth_user(self, username, password):
        pass


class EmptyAuthenticator(Authenticator):
    def auth_user(self, username, password):
        return None



import uuid

from flask_jwt_simple import get_jwt_identity
from flask_restplus import Namespace, fields, Resource, reqparse

from . import api, status_model, delete_kubernetes_deployment
from rekcurd_dashboard.models import db, ApplicationModel, ApplicationUserRoleModel, ApplicationRole, KubernetesModel, ServiceModel
from rekcurd_dashboard.utils import RekcurdDashboardException
from rekcurd_dashboard.apis import DatetimeToTimestamp


application_api_namespace = Namespace('applications', description='Application API Endpoint.')
success_or_not = application_api_namespace.model('Success', status_model)
application_model_params = application_api_namespace.model('Application', {
    'application_id': fields.String(
        readOnly=True,
        description='Application ID.'
    ),
    'application_name': fields.String(
        required=True,
        description='Application name.',
        example='rekcurd-sample'
    ),
    'project_id': fields.Integer(
        required=False,
        description='Project ID.'
    ),
    'description': fields.String(
        required=False,
        description='Description.',
        example='This is a sample.'
    ),
    'register_date': DatetimeToTimestamp(
        readOnly=True,
        description='Register date.'
    )
})


@application_api_namespace.route('/projects/<int:project_id>/applications')
class ApiApplications(Resource):
    add_application_parser = reqparse.RequestParser()
    add_application_parser.add_argument('application_name', type=str, required=True, location='form')
    add_application_parser.add_argument('description', type=str, required=False, location='form')

    @application_api_namespace.marshal_list_with(application_model_params)
    def get(self, project_id: int):
        """get_applications"""
        return ApplicationModel.query.filter_by(project_id=project_id).all()

    @application_api_namespace.marshal_with(success_or_not)
    @application_api_namespace.expect(add_application_parser)
    def post(self, project_id: int):
        """add_application"""
        args = self.add_application_parser.parse_args()
        application_name = args['application_name']
        description = args['description']

        application_model = db.session.query(ApplicationModel).filter(
            ApplicationModel.project_id == project_id,
            ApplicationModel.application_name == application_name).one_or_none()
        if application_model is not None:
            raise RekcurdDashboardException("Application name is duplicated.")
        application_id = uuid.uuid4().hex
        application_model = ApplicationModel(
            project_id=project_id, application_id=application_id,
            application_name=application_name, description=description)
        db.session.add(application_model)
        db.session.flush()

        if api.dashboard_config.IS_ACTIVATE_AUTH:
            user_id = get_jwt_identity()
            role = db.session.query(ApplicationUserRoleModel).filter(
                ApplicationUserRoleModel.application_id == application_model.application_id,
                ApplicationUserRoleModel.user_id == user_id).one_or_none()
            if role is None:
                role = ApplicationUserRoleModel(
                    application_id=application_model.application_id,
                    user_id=user_id,
                    application_role=ApplicationRole.admin.name)
                db.session.add(role)
                db.session.flush()

        db.session.commit()
        db.session.close()
        return {"status": True, "message": "Success."}


@application_api_namespace.route('/projects/<int:project_id>/applications/<application_id>')
class ApiApplicationId(Resource):
    edit_application_parser = reqparse.RequestParser()
    edit_application_parser.add_argument('description', type=str, required=True, location='form')

    @application_api_namespace.marshal_with(application_model_params)
    def get(self, project_id: int, application_id: str):
        """get_application"""
        return ApplicationModel.query.filter_by(application_id=application_id).first_or_404()

    @application_api_namespace.marshal_with(success_or_not)
    @application_api_namespace.expect(edit_application_parser)
    def patch(self, project_id: int, application_id: str):
        """update_application"""
        args = self.edit_application_parser.parse_args()
        description = args['description']
        application_model = db.session.query(ApplicationModel).filter(
            ApplicationModel.application_id == application_id).first_or_404()
        application_model.description = description
        db.session.commit()
        db.session.close()
        return {"status": True, "message": "Success."}

    @application_api_namespace.marshal_with(success_or_not)
    def delete(self, project_id: int, application_id: str):
        """delete_application"""
        kubernetes_models = db.session.query(KubernetesModel).filter(
            KubernetesModel.project_id == project_id).all()
        application_model = db.session.query(ApplicationModel).filter(
            ApplicationModel.application_id == application_id).first_or_404()
        service_models = db.session.query(ServiceModel).filter(
            ServiceModel.application_id == application_id).all()
        if len(kubernetes_models):
            """If Kubernetes mode, then request deletion to Kubernetes WebAPI"""
            for service_model in service_models:
                delete_kubernetes_deployment(kubernetes_models, application_id, service_model.service_id)
            db.session.flush()
        else:
            """Otherwise, delete DB entry."""
            # TODO: Kill service process.
            pass
        db.session.delete(application_model)
        db.session.commit()
        db.session.close()
        return {"status": True, "message": "Success."}



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# DO NOT EDIT HERE!!

import traceback, types
import grpc

from rekcurd_dashboard.protobuf import rekcurd_pb2, rekcurd_pb2_grpc

from rekcurd_dashboard.logger import SystemLoggerInterface, JsonSystemLogger
from rekcurd_dashboard.utils import ProtobufUtil
from werkzeug.datastructures import FileStorage
from protobuf_to_dict import protobuf_to_dict


def error_handling(error_response):
    """ Decorator for handling error

    Apply following processing on Servicer methods
    to handle errors.

    - Call :func:``on_error`` method (if defined) in the class
      to postprocess something on error

    Parameters
    ----------
    error_response
    """
    def _wrapper_maker(func):
        def _wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as error:
                servicer = args[0]
                if hasattr(servicer, 'on_error'):
                    assert isinstance(servicer.on_error, types.MethodType), \
                        'You must define on_error as method'
                    servicer.on_error(error)
                return error_response
        return _wrapper
    return _wrapper_maker


class RekcurdDashboardClient:
    _logger: SystemLoggerInterface = None

    def __init__(self, host: str = None, port: int = None,
                 application_name: str = None, service_level: str = None,
                 rekcurd_grpc_version: str = None):
        self.logger = JsonSystemLogger()

        _host = "127.0.0.1"
        _port = 5000
        host = host or _host
        port = int(port or _port)

        if rekcurd_grpc_version is None:
            rekcurd_grpc_version = rekcurd_pb2.DESCRIPTOR.GetOptions().Extensions[rekcurd_pb2.rekcurd_grpc_proto_version]
        else:
            rekcurd_pb2.EnumVersionInfo.Value(rekcurd_grpc_version)

        self.__metadata = [('x-rekcurd-application-name', application_name),
                           ('x-rekcurd-sevice-level', service_level),
                           ('x-rekcurd-grpc-version', rekcurd_grpc_version)]

        channel = grpc.insecure_channel("{}:{}".format(host, port))
        self.stub = rekcurd_pb2_grpc.RekcurdDashboardStub(channel)

    @property
    def logger(self):
        return self._logger

    @logger.setter
    def logger(self, logger: SystemLoggerInterface):
        if isinstance(logger, SystemLoggerInterface):
            self._logger = logger
        else:
            raise TypeError("Invalid logger type.")

    def on_error(self, error: Exception):
        """ Postprocessing on error

        For detail, see :func:``on_error``

        Parameters
        ----------
        error : Exception
            Error to be handled
        """
        self.logger.error(str(error))
        self.logger.error(traceback.format_exc())

    @error_handling({"status": False})
    def run_service_info(self):
        request = rekcurd_pb2.ServiceInfoRequest()
        response_protobuf = self.stub.ServiceInfo(request, metadata=self.__metadata)
        response = protobuf_to_dict(response_protobuf,
                                    including_default_value_fields=True)
        return response

    def __upload_model_request(self, model_path:str, f:FileStorage):
        for data in ProtobufUtil.stream_file(f):
            request = rekcurd_pb2.UploadModelRequest(
                path=model_path,
                data=data
            )
            yield request

    @error_handling({"status": False})
    def run_upload_model(self, model_path:str, f:FileStorage):
        request_iterator = self.__upload_model_request(model_path, f)
        response = self.stub.UploadModel(request_iterator, metadata=self.__metadata)
        response = protobuf_to_dict(response,
                                    including_default_value_fields=True)
        if not response["status"]:
            raise Exception(response["message"])
        return response

    @error_handling({"status": False})
    def run_switch_service_model_assignment(self, model_path:str):
        request = rekcurd_pb2.SwitchModelRequest(
            path=model_path,
        )
        response_protobuf = self.stub.SwitchModel(request, metadata=self.__metadata)
        response = protobuf_to_dict(response_protobuf,
                                    including_default_value_fields=True)
        if not response["status"]:
            raise Exception(response["message"])
        return response

    @error_handling({"status": False})
    def run_evaluate_model(self, data_path:str, result_path:str):
        request_iterator = iter([rekcurd_pb2.EvaluateModelRequest(data_path=data_path, result_path=result_path)])
        metrics = self.stub.EvaluateModel(request_iterator, metadata=self.__metadata).metrics
        response = dict(protobuf_to_dict(metrics, including_default_value_fields=True),
                        label=[self.__get_value_from_io(l) for l in metrics.label])
        response['status'] = True
        return response

    def __upload_eval_data_request(self, f:FileStorage, data_path:str):
        for data in ProtobufUtil.stream_file(f):
            request = rekcurd_pb2.UploadEvaluationDataRequest(data=data, data_path=data_path)
            yield request

    @error_handling({"status": False})
    def run_upload_evaluation_data(self, f:FileStorage, data_path:str):
        request_iterator = self.__upload_eval_data_request(f, data_path)
        response = protobuf_to_dict(self.stub.UploadEvaluationData(request_iterator, metadata=self.__metadata),
                                    including_default_value_fields=True)
        return response

    def __get_value_from_io(self, io:rekcurd_pb2.IO):
        if io.WhichOneof('io_oneof') == 'str':
            val = io.str.val
        else:
            val = io.tensor.val

        if len(val) == 1:
            return val[0]
        else:
            return list(val)

    @error_handling({"status": False})
    def run_evaluation_data(self, data_path:str, result_path:str):
        request = rekcurd_pb2.EvaluationResultRequest(data_path=data_path, result_path=result_path)
        for raw_response in self.stub.EvaluationResult(request, metadata=self.__metadata):
            details = []
            for detail in raw_response.detail:
                details.append(dict(
                    protobuf_to_dict(detail, including_default_value_fields=True),
                    input=self.__get_value_from_io(detail.input),
                    label=self.__get_value_from_io(detail.label),
                    output=self.__get_value_from_io(detail.output),
                    score=detail.score[0] if len(detail.score) == 1 else list(detail.score)
                ))
            metrics = raw_response.metrics
            metrics_response = dict(protobuf_to_dict(metrics, including_default_value_fields=True),
                                    label=[self.__get_value_from_io(l) for l in metrics.label])
            response = protobuf_to_dict(raw_response,
                                        including_default_value_fields=True)
            response['detail'] = details
            response['status'] = True
            response['metrics'] = metrics_response
            yield response



from flask_restplus import Namespace, fields, Resource, reqparse, inputs

from . import status_model
from rekcurd_dashboard.models import db, DataServerModel, DataServerModeEnum
from rekcurd_dashboard.utils import RekcurdDashboardException
from rekcurd_dashboard.apis import DatetimeToTimestamp


data_server_api_namespace = Namespace('dataservers', description='Data Server API Endpoint.')
success_or_not = data_server_api_namespace.model('Success', status_model)
data_server_model_params = data_server_api_namespace.model('DataServer', {
    'project_id': fields.Integer(
        required=True,
        description='Project ID.'
    ),
    'data_server_mode': fields.String(
        required=True,
        description='Data server mode. [local/ceph_s3/aws_s3]',
    ),
    'ceph_access_key': fields.String(
        required=False,
        description='Ceph S3 API access key.'
    ),
    'ceph_secret_key': fields.String(
        required=False,
        description='Ceph S3 API secret key.'
    ),
    'ceph_host': fields.String(
        required=False,
        description='Ceph S3 API endpoint host name.'
    ),
    'ceph_port': fields.Integer(
        required=False,
        description='Ceph S3 API port number.'
    ),
    'ceph_is_secure': fields.Boolean(
        required=False,
        description='Ceph S3 API is SSL or not.'
    ),
    'ceph_bucket_name': fields.String(
        required=False,
        description='Ceph S3 API bucket name.'
    ),
    'aws_access_key': fields.String(
        required=False,
        description='AWS S3 API access key.'
    ),
    'aws_secret_key': fields.String(
        required=False,
        description='AWS S3 API secret key.'
    ),
    'aws_bucket_name': fields.String(
        required=False,
        description='AWS S3 API bucket name.'
    ),
    'register_date': DatetimeToTimestamp(
        readOnly=True,
        description='Register date.'
    )
})

data_server_parser = reqparse.RequestParser()
data_server_parser.add_argument(
    'data_server_mode', location='form', type=str, required=True,
    choices=('local', 'ceph_s3', 'aws_s3'),
    help='Data server mode. [local/ceph_s3/aws_s3].')
data_server_parser.add_argument(
    'ceph_access_key', location='form', type=str, required=False,
    default='',
    help='Ceph S3 API access key.')
data_server_parser.add_argument(
    'ceph_secret_key', location='form', type=str, required=False,
    default='',
    help='Ceph S3 API secret key.')
data_server_parser.add_argument(
    'ceph_host', location='form', type=str, required=False,
    default='',
    help='Ceph S3 API endpoint host name.')
data_server_parser.add_argument(
    'ceph_port', location='form', type=int, required=False,
    default=80,
    help='Ceph S3 API port number.')
data_server_parser.add_argument(
    'ceph_is_secure', location='form', type=inputs.boolean, required=False,
    default=False,
    help='Ceph S3 API is SSL or not.')
data_server_parser.add_argument(
    'ceph_bucket_name', location='form', type=str, required=False,
    default='',
    help='Ceph S3 API bucket name.')
data_server_parser.add_argument(
    'aws_access_key', location='form', type=str, required=False,
    default='',
    help='AWS S3 API access key.')
data_server_parser.add_argument(
    'aws_secret_key', location='form', type=str, required=False,
    default='',
    help='AWS S3 API secret key.')
data_server_parser.add_argument(
    'aws_bucket_name', location='form', type=str, required=False,
    default='',
    help='AWS S3 API bucket name.')


@data_server_api_namespace.route('/projects/<int:project_id>/data_servers')
class ApiDataServers(Resource):
    @data_server_api_namespace.marshal_with(data_server_model_params)
    def get(self, project_id: int):
        """get_data_servers"""
        return DataServerModel.query.filter_by(project_id=project_id).first_or_404()

    @data_server_api_namespace.marshal_with(success_or_not)
    @data_server_api_namespace.expect(data_server_parser)
    def post(self, project_id: int):
        """add_data_server"""
        args = data_server_parser.parse_args()
        data_server_mode = args['data_server_mode']
        data_server_mode_enum = DataServerModeEnum.to_enum(data_server_mode)
        ceph_access_key = args['ceph_access_key']
        ceph_secret_key = args['ceph_secret_key']
        ceph_host = args['ceph_host']
        ceph_port = args['ceph_port']
        ceph_is_secure = args['ceph_is_secure']
        ceph_bucket_name = args['ceph_bucket_name']
        aws_access_key = args['aws_access_key']
        aws_secret_key = args['aws_secret_key']
        aws_bucket_name = args['aws_bucket_name']
        if data_server_mode_enum == DataServerModeEnum.LOCAL:
            pass
        elif data_server_mode_enum == DataServerModeEnum.CEPH_S3:
            if ceph_access_key and ceph_secret_key and ceph_host and \
                    ceph_port and ceph_is_secure is not None and ceph_bucket_name:
                pass
            else:
                raise RekcurdDashboardException(
                    "Need to set \"ceph_access_key\", \"ceph_secret_key\", \"ceph_host\", \"ceph_port\", "
                    "\"ceph_is_secure\" and \"ceph_bucket_name\"")
        elif data_server_mode_enum == DataServerModeEnum.AWS_S3:
            if aws_bucket_name:
                pass
            else:
                raise RekcurdDashboardException(
                    "Need to set \"aws_bucket_name\"")
        else:
            raise RekcurdDashboardException("Invalid data server mode specified.")

        data_server_model = db.session.query(DataServerModel).filter(
            DataServerModel.project_id == project_id).one_or_none()
        if data_server_model is not None:
            raise RekcurdDashboardException("Data server exists already.")
        data_server_model = DataServerModel(
            project_id=project_id, data_server_mode=data_server_mode_enum,
            ceph_access_key=ceph_access_key, ceph_secret_key=ceph_secret_key,
            ceph_host=ceph_host, ceph_port=ceph_port, ceph_is_secure=ceph_is_secure,
            ceph_bucket_name=ceph_bucket_name, aws_access_key=aws_access_key,
            aws_secret_key=aws_secret_key, aws_bucket_name=aws_bucket_name)
        db.session.add(data_server_model)
        db.session.commit()
        db.session.close()
        return {"status": True, "message": "Success."}

    @data_server_api_namespace.marshal_with(success_or_not)
    @data_server_api_namespace.expect(data_server_parser)
    def patch(self, project_id: int):
        """update_data_server"""
        args = data_server_parser.parse_args()
        data_server_mode = args['data_server_mode']
        data_server_mode_enum = DataServerModeEnum.to_enum(data_server_mode)
        ceph_access_key = args['ceph_access_key']
        ceph_secret_key = args['ceph_secret_key']
        ceph_host = args['ceph_host']
        ceph_port = args['ceph_port']
        ceph_is_secure = args['ceph_is_secure']
        ceph_bucket_name = args['ceph_bucket_name']
        aws_access_key = args['aws_access_key']
        aws_secret_key = args['aws_secret_key']
        aws_bucket_name = args['aws_bucket_name']

        data_server_model: DataServerModel = db.session.query(DataServerModel).filter(
            DataServerModel.project_id == project_id).first_or_404()
        is_updated = False
        if data_server_model.data_server_mode != data_server_mode_enum:
            is_updated = True
            data_server_model.data_server_mode = data_server_mode_enum
        if ceph_access_key is not None:
            is_updated = True
            data_server_model.ceph_access_key = ceph_access_key
        if ceph_secret_key is not None:
            is_updated = True
            data_server_model.ceph_secret_key = ceph_secret_key
        if ceph_host is not None:
            is_updated = True
            data_server_model.ceph_host = ceph_host
        if ceph_port is not None:
            is_updated = True
            data_server_model.ceph_port = ceph_port
        if ceph_is_secure is not None:
            is_updated = True
            data_server_model.ceph_is_secure = ceph_is_secure
        if ceph_bucket_name is not None:
            is_updated = True
            data_server_model.ceph_bucket_name = ceph_bucket_name
        if aws_access_key is not None:
            is_updated = True
            data_server_model.aws_access_key = aws_access_key
        if aws_secret_key is not None:
            is_updated = True
            data_server_model.aws_secret_key = aws_secret_key
        if aws_bucket_name is not None:
            is_updated = True
            data_server_model.aws_bucket_name = aws_bucket_name

        if data_server_mode_enum == DataServerModeEnum.LOCAL:
            pass
        elif data_server_mode_enum == DataServerModeEnum.CEPH_S3:
            if data_server_model.ceph_access_key and data_server_model.ceph_secret_key and \
                    data_server_model.ceph_host and data_server_model.ceph_port and \
                    data_server_model.ceph_is_secure is not None and data_server_model.ceph_bucket_name:
                pass
            else:
                raise RekcurdDashboardException(
                    "Need to set \"ceph_access_key\", \"ceph_secret_key\", \"ceph_host\", \"ceph_port\", "
                    "\"ceph_is_secure\" and \"ceph_bucket_name\"")
        elif data_server_mode_enum == DataServerModeEnum.AWS_S3:
            if data_server_model.aws_bucket_name:
                pass
            else:
                raise RekcurdDashboardException(
                    "Need to set \"aws_bucket_name\"")
        else:
            raise RekcurdDashboardException("Invalid data server mode specified.")

        if is_updated:
            db.session.commit()
        db.session.close()
        return {"status": True, "message": "Success."}

    @data_server_api_namespace.marshal_with(success_or_not)
    def delete(self, project_id: int):
        """delete_data_server"""
        db.session.query(DataServerModel).filter(DataServerModel.project_id == project_id).delete()
        db.session.commit()
        db.session.close()
        return {"status": True, "message": "Success."}
